# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import logging
import os
import socket
import sys

import netaddr
import netifaces
import requests
import werkzeug.http

_cached_check_v6 = None


def check_v6():
    global _cached_check_v6

    def f():
        if not socket.has_ipv6:
            return False

        try:
            socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        except Exception:
            # "[Errno 97] Address family not supported by protocol" or anything else really...
            return False
        return True

    if _cached_check_v6 is None:
        _cached_check_v6 = f()
    return _cached_check_v6


HAS_V6 = check_v6()

if hasattr(socket, "IPPROTO_IPV6") and hasattr(socket, "IPV6_V6ONLY"):
    # Dual stack support, hooray!
    IPPROTO_IPV6 = socket.IPPROTO_IPV6
    IPV6_V6ONLY = socket.IPV6_V6ONLY
else:
    if sys.platform == "win32":
        # Python 2.7 on Windows lacks IPPROTO_IPV6, but supports the socket options just fine, let's redefine it
        IPPROTO_IPV6 = 41
        IPV6_V6ONLY = 27
    else:
        # Whatever we are running on here, we don't want to use V6 on here due to lack of dual stack support
        HAS_V6 = False


def get_lan_ranges(additional_private=None):
    logger = logging.getLogger(__name__)

    if additional_private is None or not isinstance(additional_private, (list, tuple)):
        additional_private = []

    def to_ipnetwork(address):
        prefix = address["netmask"]
        if "/" in prefix:
            # v6 notation in netifaces output, e.g. "ffff:ffff:ffff:ffff::/64"
            _, prefix = prefix.split("/")

        addr = strip_interface_tag(address["addr"])
        return netaddr.IPNetwork("{}/{}".format(addr, prefix))

    subnets = []

    for interface in netifaces.interfaces():
        addrs = netifaces.ifaddresses(interface)
        for v4 in addrs.get(socket.AF_INET, ()):
            try:
                subnets.append(to_ipnetwork(v4))
            except Exception:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception(
                        "Error while trying to add v4 network to local subnets: {!r}".format(
                            v4
                        )
                    )

        if HAS_V6:
            for v6 in addrs.get(socket.AF_INET6, ()):
                try:
                    subnets.append(to_ipnetwork(v6))
                except Exception:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.exception(
                            "Error while trying to add v6 network to local subnets: {!r}".format(
                                v6
                            )
                        )

    for additional in additional_private:
        try:
            subnets.append(netaddr.IPNetwork(additional))
        except Exception:
            if logger.isEnabledFor(logging.DEBUG):
                logger.exception(
                    "Error while trying to add additional private network to local subnets: {}".format(
                        additional
                    )
                )

    subnets += list(netaddr.ip.IPV4_PRIVATE) + [
        netaddr.ip.IPV4_LOOPBACK,
        netaddr.ip.IPV4_LINK_LOCAL,
    ]
    if HAS_V6:
        subnets += list(netaddr.ip.IPV6_PRIVATE) + [
            netaddr.IPNetwork(netaddr.ip.IPV6_LOOPBACK),
            netaddr.ip.IPV6_LINK_LOCAL,
        ]

    return subnets


def is_lan_address(address, additional_private=None):
    if not address:
        # no address is LAN address
        return True

    try:
        address = sanitize_address(address)
        ip = netaddr.IPAddress(address)
        subnets = get_lan_ranges(additional_private=additional_private)

        if any(map(lambda subnet: ip in subnet, subnets)):
            return True

        return False

    except Exception:
        # we are extra careful here since an unhandled exception in this method will effectively nuke the whole UI
        logging.getLogger(__name__).exception(
            "Error while trying to determine whether {} is a local address".format(
                address
            )
        )
        return True


def sanitize_address(address):
    address = unmap_v4_as_v6(address)
    address = strip_interface_tag(address)
    return address


def strip_interface_tag(address):
    if "%" in address:
        # interface comment, e.g. "fe80::457f:bbee:d579:1063%wlan0"
        address = address[: address.find("%")]
    return address


def unmap_v4_as_v6(address):
    if address.lower().startswith("::ffff:") and "." in address:
        # ipv6 mapped ipv4 address, unmap
        address = address[len("::ffff:") :]
    return address


def interface_addresses(family=None, interfaces=None):
    """
    Retrieves all of the host's network interface addresses.
    """

    import netifaces

    if not family:
        family = netifaces.AF_INET

    if interfaces is None:
        interfaces = netifaces.interfaces()

    for interface in interfaces:
        try:
            ifaddresses = netifaces.ifaddresses(interface)
        except Exception:
            continue
        if family in ifaddresses:
            for ifaddress in ifaddresses[family]:
                address = netaddr.IPAddress(ifaddress["addr"])
                if not address.is_link_local() and not address.is_loopback():
                    yield ifaddress["addr"]


def address_for_client(host, port, timeout=3.05, addresses=None, interfaces=None):
    """
    Determines the address of the network interface on this host needed to connect to the indicated client host and port.
    """

    if addresses is None:
        addresses = interface_addresses(interfaces=interfaces)

    for address in addresses:
        try:
            if server_reachable(host, port, timeout=timeout, proto="udp", source=address):
                return address
        except Exception:
            continue


def server_reachable(host, port, timeout=3.05, proto="tcp", source=None):
    """
    Checks if a server is reachable

    Args:
            host (str): host to check against
            port (int): port to check against
            timeout (float): timeout for check
            proto (str): ``tcp`` or ``udp``
            source (str): optional, socket used for check will be bound against this address if provided

    Returns:
            boolean: True if a connection to the server could be opened, False otherwise
    """

    import socket

    if proto not in ("tcp", "udp"):
        raise ValueError("proto must be either 'tcp' or 'udp'")

    try:
        sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM if proto == "udp" else socket.SOCK_STREAM
        )
        sock.settimeout(timeout)
        if source is not None:
            sock.bind((source, 0))
        sock.connect((host, port))
        return True
    except Exception:
        return False


def resolve_host(host):
    import socket

    from octoprint.util import to_unicode

    try:
        return [to_unicode(x[4][0]) for x in socket.getaddrinfo(host, 80)]
    except Exception:
        return []


def download_file(url, folder, max_length=None):
    with requests.get(url, stream=True) as r:
        r.raise_for_status()

        filename = None
        if "Content-Disposition" in r.headers.keys():
            _, options = werkzeug.http.parse_options_header(
                r.headers["Content-Disposition"]
            )
            filename = options.get("filename")
        if filename is None:
            filename = url.split("/")[-1]

        assert len(filename) > 0

        # TODO check content-length against safety limit

        path = os.path.abspath(os.path.join(folder, filename))
        assert path.startswith(folder)

        with io.open(path, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return path



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

"""
In this module resides the core data structures and logic of the plugin system.

.. autoclass:: PluginManager
   :members:

.. autoclass:: PluginInfo
   :members:

.. autoclass:: Plugin
   :members:

.. autoclass:: RestartNeedingPlugin
   :members:

.. autoclass:: SortablePlugin
   :members:

"""

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"


import fnmatch
import inspect
import io
import logging
import os
import sys
from collections import OrderedDict, defaultdict, namedtuple

import pkg_resources
import pkginfo
from past.builtins import unicode

from octoprint.util import sv, time_this, to_unicode
from octoprint.util.version import get_python_version_string, is_python_compatible

try:
    from os import scandir
except ImportError:
    from scandir import scandir


if sys.version_info[0] == 2:
    # noinspection PyDeprecation
    import imp
else:
    # deprecated in Python 3.4+ and hence vendored for now
    import octoprint.vendor.imp as imp


# noinspection PyDeprecation
def _find_module(name, path=None):
    if path is not None:
        spec = imp.find_module(name, [path])
    else:
        spec = imp.find_module(name)
    return spec[1], spec


# noinspection PyDeprecation
def _load_module(name, spec):
    f, filename, details = spec
    return imp.load_module(name, f, filename, details)


_EntryPointOrigin = namedtuple(
    "EntryPointOrigin", "type, entry_point, module_name, package_name, package_version"
)


class EntryPointOrigin(_EntryPointOrigin):
    """
    Origin of a plugin registered via an entry point.

    .. attribute:: type

       Always ``entry_point``.

    .. attribute:: entry_point

       Name of the entry point, usually ``octoprint.plugin``.

    .. attribute:: module_name

       Module registered to the entry point.

    .. attribute:: package_name

       Python package containing the entry point.

    .. attribute:: package_version

       Version of the python package containing the entry point.
    """


_FolderOrigin = namedtuple("FolderOrigin", "type, folder")


class FolderOrigin(_FolderOrigin):
    """
    Origin of a (single file) plugin loaded from a plugin folder.

    .. attribute:: type

       Always `folder`.

    .. attribute:: folder

       Folder path from which the plugin was loaded.
    """


_ModuleOrigin = namedtuple("ModuleOrigin", "type, module_name, folder")


class ModuleOrigin(_ModuleOrigin):
    """
    Origin of a (single file) plugin loaded from a plugin folder.

    .. attribute:: type

       Always `module`.

    .. attribute:: module_name

       Name of the module from which the plugin was loaded.

    .. attribute:: folder

       Folder path from which the plugin was loaded.
    """


class PluginInfo(object):
    """
    The :class:`PluginInfo` class wraps all available information about a registered plugin.

    This includes its meta data (like name, description, version, etc) as well as the actual plugin extensions like
    implementations, hooks and helpers.

    It works on Python module objects and extracts the relevant data from those via accessing the
    :ref:`control properties <sec-plugins-controlproperties>`.

    Arguments:
        key (str): Identifier of the plugin
        location (str): Installation folder of the plugin
        instance (module): Plugin module instance - this may be ``None`` if the plugin has been blacklisted!
        name (str): Human readable name of the plugin
        version (str): Version of the plugin
        description (str): Description of the plugin
        author (str): Author of the plugin
        url (str): URL of the website of the plugin
        license (str): License of the plugin
    """

    attr_name = "__plugin_name__"
    """ Module attribute from which to retrieve the plugin's human readable name. """

    attr_description = "__plugin_description__"
    """ Module attribute from which to retrieve the plugin's description. """

    attr_disabling_discouraged = "__plugin_disabling_discouraged__"
    """ Module attribute from which to retrieve the reason why disabling the plugin is discouraged. Only effective if ``self.bundled`` is True. """

    attr_version = "__plugin_version__"
    """ Module attribute from which to retrieve the plugin's version. """

    attr_author = "__plugin_author__"
    """ Module attribute from which to retrieve the plugin's author. """

    attr_url = "__plugin_url__"
    """ Module attribute from which to retrieve the plugin's website URL. """

    attr_license = "__plugin_license__"
    """ Module attribute from which to retrieve the plugin's license. """

    attr_pythoncompat = "__plugin_pythoncompat__"
    """
    Module attribute from which to retrieve the plugin's python compatibility string.

    If unset a default of ``>=2.7,<3`` will be assumed, meaning that the plugin will be considered compatible to
    Python 2 but not Python 3.

    To mark a plugin as Python 3 compatible, a string of ``>=2.7,<4`` is recommended.

    Bundled plugins will automatically be assumed to be compatible.
    """

    attr_hidden = "__plugin_hidden__"
    """
    Module attribute from which to determine if the plugin's hidden or not.

    Only evaluated for bundled plugins, in order to hide them from the Plugin Manager
    and similar places.
    """

    attr_hooks = "__plugin_hooks__"
    """ Module attribute from which to retrieve the plugin's provided hooks. """

    attr_implementation = "__plugin_implementation__"
    """ Module attribute from which to retrieve the plugin's provided mixin implementation. """

    attr_implementations = "__plugin_implementations__"
    """
    Module attribute from which to retrieve the plugin's provided implementations.

    This deprecated attribute will only be used if a plugin does not yet offer :attr:`attr_implementation`. Only the
    first entry will be evaluated.

    .. deprecated:: 1.2.0-dev-694

       Use :attr:`attr_implementation` instead.
    """

    attr_helpers = "__plugin_helpers__"
    """ Module attribute from which to retrieve the plugin's provided helpers. """

    attr_check = "__plugin_check__"
    """ Module attribute which to call to determine if the plugin can be loaded. """

    attr_init = "__plugin_init__"
    """
    Module attribute which to call when loading the plugin.

    This deprecated attribute will only be used if a plugin does not yet offer :attr:`attr_load`.

    .. deprecated:: 1.2.0-dev-720

       Use :attr:`attr_load` instead.
    """

    attr_load = "__plugin_load__"
    """ Module attribute which to call when loading the plugin. """

    attr_unload = "__plugin_unload__"
    """ Module attribute which to call when unloading the plugin. """

    attr_enable = "__plugin_enable__"
    """ Module attribute which to call when enabling the plugin. """

    attr_disable = "__plugin_disable__"
    """ Module attribute which to call when disabling the plugin. """

    def __init__(
        self,
        key,
        location,
        instance,
        name=None,
        version=None,
        description=None,
        author=None,
        url=None,
        license=None,
        parsed_metadata=None,
    ):
        self.key = key
        self.location = location
        self.instance = instance

        self.origin = None
        """
        The origin from which this plugin was loaded, either a :class:`EntryPointOrigin`, :class:`FolderOrigin`
        or :class:`ModuleOrigin` instance. Set during loading, initially ``None``.
        """

        self.enabled = True
        """Whether the plugin is enabled."""

        self.blacklisted = False
        """Whether the plugin is blacklisted."""

        self.forced_disabled = False
        """Whether the plugin has been force disabled by the system, e.g. due to safe mode blacklisting."""

        self.incompatible = False
        """Whether this plugin has been detected as incompatible."""

        self.bundled = False
        """Whether this plugin is bundled with OctoPrint."""

        self.loaded = False
        """Whether this plugin has been loaded."""

        self.managable = True
        """Whether this plugin can be managed by OctoPrint."""

        self.needs_restart = False
        """Whether this plugin needs a restart of OctoPrint after enabling/disabling."""

        self.invalid_syntax = False
        """Whether invalid syntax was encountered while trying to load this plugin."""

        self._name = name
        self._version = version
        self._description = description
        self._author = author
        self._url = url
        self._license = license

        self._logger = logging.getLogger(__name__)

        self._cached_parsed_metadata = parsed_metadata
        if self._cached_parsed_metadata is None:
            self._cached_parsed_metadata = self._parse_metadata()

    def validate(self, phase, additional_validators=None):
        """
        Validates the plugin for various validation phases.

        ``phase`` can be one of ``before_import``, ``before_load``, ``after_load``.

        Used by :class:`PluginManager`, should not be used elsewhere.
        """
        result = True

        if phase == "before_import":
            result = (
                not self.forced_disabled
                and not self.blacklisted
                and not self.incompatible
                and not self.invalid_syntax
                and result
            )

        elif phase == "before_load":
            # if the plugin still uses __plugin_init__, log a deprecation warning and move it to __plugin_load__
            if hasattr(self.instance, self.__class__.attr_init):
                if not hasattr(self.instance, self.__class__.attr_load):
                    # deprecation warning
                    import warnings

                    warnings.warn(
                        "{name} uses deprecated control property __plugin_init__, use __plugin_load__ instead".format(
                            name=self.key
                        ),
                        DeprecationWarning,
                    )

                    # move it
                    init = getattr(self.instance, self.__class__.attr_init)
                    setattr(self.instance, self.__class__.attr_load, init)

                # delete __plugin_init__
                delattr(self.instance, self.__class__.attr_init)

        elif phase == "after_load":
            # if the plugin still uses __plugin_implementations__, log a deprecation warning and put the first
            # item into __plugin_implementation__
            if hasattr(self.instance, self.__class__.attr_implementations):
                if not hasattr(self.instance, self.__class__.attr_implementation):
                    # deprecation warning
                    import warnings

                    warnings.warn(
                        "{name} uses deprecated control property __plugin_implementations__, use __plugin_implementation__ instead - only the first implementation of {name} will be recognized".format(
                            name=self.key
                        ),
                        DeprecationWarning,
                    )

                    # put first item into __plugin_implementation__
                    implementations = getattr(
                        self.instance, self.__class__.attr_implementations
                    )
                    if len(implementations) > 0:
                        setattr(
                            self.instance,
                            self.__class__.attr_implementation,
                            implementations[0],
                        )

                # delete __plugin_implementations__
                delattr(self.instance, self.__class__.attr_implementations)

        if additional_validators is not None:
            for validator in additional_validators:
                result = validator(phase, self) and result

        return result

    def __str__(self):
        if self.version:
            return "{name} ({version})".format(name=self.name, version=self.version)
        else:
            return to_unicode(self.name)

    def long_str(
        self,
        show_bundled=False,
        bundled_strs=(" [B]", ""),
        show_location=False,
        location_str=" - {location}",
        show_enabled=False,
        enabled_strs=("* ", "  ", "X ", "C "),
    ):
        """
        Long string representation of the plugin's information. Will return a string of the format ``<enabled><str(self)><bundled><location>``.

        ``enabled``, ``bundled`` and ``location`` will only be displayed if the corresponding flags are set to ``True``.
        The will be filled from ``enabled_str``, ``bundled_str`` and ``location_str`` as follows:

        ``enabled_str``
            a 4-tuple, the first entry being the string to insert when the plugin is enabled, the second
            entry the string to insert when it is not, the third entry the string when it is blacklisted
            and the fourth when it is incompatible.
        ``bundled_str``
            a 2-tuple, the first entry being the string to insert when the plugin is bundled, the second
            entry the string to insert when it is not.
        ``location_str``
            a format string (to be parsed with ``str.format``), the ``{location}`` placeholder will be
            replaced with the plugin's installation folder on disk.

        Arguments:
            show_enabled (boolean): whether to show the ``enabled`` part
            enabled_strs (tuple): the 2-tuple containing the two possible strings to use for displaying the enabled state
            show_bundled (boolean): whether to show the ``bundled`` part
            bundled_strs(tuple): the 2-tuple containing the two possible strings to use for displaying the bundled state
            show_location (boolean): whether to show the ``location`` part
            location_str (str): the format string to use for displaying the plugin's installation location

        Returns:
            str: The long string representation of the plugin as described above
        """
        if show_enabled:
            if self.incompatible:
                ret = to_unicode(enabled_strs[3])
            elif self.blacklisted:
                ret = to_unicode(enabled_strs[2])
            elif not self.enabled:
                ret = to_unicode(enabled_strs[1])
            else:
                ret = to_unicode(enabled_strs[0])
        else:
            ret = ""

        ret += unicode(self)

        if show_bundled:
            ret += (
                to_unicode(bundled_strs[0])
                if self.bundled
                else to_unicode(bundled_strs[1])
            )

        if show_location and self.location:
            ret += to_unicode(location_str).format(location=self.location)

        return ret

    def get_hook(self, hook):
        """
        Arguments:
            hook (str): Hook to return.

        Returns:
            callable or None: Handler for the requested ``hook`` or None if no handler is registered.
        """

        if hook not in self.hooks:
            return None
        return self.hooks[hook]

    def get_implementation(self, *types):
        """
        Arguments:
            types (list): List of :class:`Plugin` sub classes the implementation needs to implement.

        Returns:
            object: The plugin's implementation if it matches all of the requested ``types``, None otherwise.
        """

        if self.implementation and all(
            map(lambda t: isinstance(self.implementation, t), types)
        ):
            return self.implementation
        else:
            return None

    @property
    def name(self):
        """
        Human readable name of the plugin. Will be taken from name attribute of the plugin module if available,
        otherwise from the ``name`` supplied during construction with a fallback to ``key``.

        Returns:
            str: Name of the plugin, fallback is the plugin's identifier.
        """
        return self._get_instance_attribute(
            self.__class__.attr_name, defaults=(self._name, self.key), incl_metadata=True
        )

    @property
    def description(self):
        """
        Description of the plugin. Will be taken from the description attribute of the plugin module as defined in
        :attr:`attr_description` if available, otherwise from the ``description`` supplied during construction.
        May be None.

        Returns:
            str or None: Description of the plugin.
        """
        return self._get_instance_attribute(
            self.__class__.attr_description, default=self._description, incl_metadata=True
        )

    @property
    def disabling_discouraged(self):
        """
        Reason why disabling of this plugin is discouraged. Only evaluated for bundled plugins! Will be taken from
        the disabling_discouraged attribute of the plugin module as defined in :attr:`attr_disabling_discouraged` if
        available. False if unset or plugin not bundled.

        Returns:
            str or None: Reason why disabling this plugin is discouraged (only for bundled plugins)
        """
        return (
            self._get_instance_attribute(
                self.__class__.attr_disabling_discouraged, default=False
            )
            if self.bundled
            else False
        )

    @property
    def version(self):
        """
        Version of the plugin. Will be taken from the version attribute of the plugin module as defined in
        :attr:`attr_version` if available, otherwise from the ``version`` supplied during construction. May be None.

        Returns:
            str or None: Version of the plugin.
        """
        return (
            self._version
            if self._version is not None
            else self._get_instance_attribute(
                self.__class__.attr_version, default=self._version, incl_metadata=True
            )
        )

    @property
    def author(self):
        """
        Author of the plugin. Will be taken from the author attribute of the plugin module as defined in
        :attr:`attr_author` if available, otherwise from the ``author`` supplied during construction. May be None.

        Returns:
            str or None: Author of the plugin.
        """
        return self._get_instance_attribute(
            self.__class__.attr_author, default=self._author, incl_metadata=True
        )

    @property
    def url(self):
        """
        Website URL for the plugin. Will be taken from the url attribute of the plugin module as defined in
        :attr:`attr_url` if available, otherwise from the ``url`` supplied during construction. May be None.

        Returns:
            str or None: Website URL for the plugin.
        """
        return self._get_instance_attribute(
            self.__class__.attr_url, default=self._url, incl_metadata=True
        )

    @property
    def license(self):
        """
        License of the plugin. Will be taken from the license attribute of the plugin module as defined in
        :attr:`attr_license` if available, otherwise from the ``license`` supplied during construction. May be None.

        Returns:
            str or None: License of the plugin.
        """
        return self._get_instance_attribute(
            self.__class__.attr_license, default=self._license, incl_metadata=True
        )

    @property
    def pythoncompat(self):
        """
        Python compatibility string of the plugin module as defined in :attr:`attr_pythoncompat` if available, otherwise
        defaults to ``>=2.7,<3``.

        Returns:
            str: Python compatibility string of the plugin
        """
        return self._get_instance_attribute(
            self.__class__.attr_pythoncompat, default=">=2.7,<3", incl_metadata=True
        )

    @property
    def hidden(self):
        """
        Hidden flag.

        Returns:
            bool: Whether the plugin should be flagged as hidden or not
        """
        return self._get_instance_attribute(
            self.__class__.attr_hidden, default=False, incl_metadata=True
        )

    @property
    def hooks(self):
        """
        Hooks provided by the plugin. Will be taken from the hooks attribute of the plugin module as defined in
        :attr:`attr_hooks` if available, otherwise an empty dictionary is returned.

        Returns:
            dict: Hooks provided by the plugin.
        """
        return self._get_instance_attribute(self.__class__.attr_hooks, default={})

    @property
    def implementation(self):
        """
        Implementation provided by the plugin. Will be taken from the implementation attribute of the plugin module
        as defined in :attr:`attr_implementation` if available, otherwise None is returned.

        Returns:
            object: Implementation provided by the plugin.
        """
        return self._get_instance_attribute(
            self.__class__.attr_implementation, default=None
        )

    @property
    def helpers(self):
        """
        Helpers provided by the plugin. Will be taken from the helpers attribute of the plugin module as defined in
        :attr:`attr_helpers` if available, otherwise an empty list is returned.

        Returns:
            dict: Helpers provided by the plugin.
        """
        return self._get_instance_attribute(self.__class__.attr_helpers, default={})

    @property
    def check(self):
        """
        Method for pre-load check of plugin. Will be taken from the check attribute of the plugin module as defined in
        :attr:`attr_check` if available, otherwise a lambda always returning True is returned.

        Returns:
            callable: Check method for the plugin module which should return True if the plugin can be loaded, False
                otherwise.
        """
        return self._get_instance_attribute(
            self.__class__.attr_check, default=lambda: True
        )

    @property
    def load(self):
        """
        Method for loading the plugin module. Will be taken from the load attribute of the plugin module as defined
        in :attr:`attr_load` if available, otherwise a no-operation lambda will be returned.

        Returns:
            callable: Load method for the plugin module.
        """
        return self._get_instance_attribute(
            self.__class__.attr_load, default=lambda: True
        )

    @property
    def unload(self):
        """
        Method for unloading the plugin module. Will be taken from the unload attribute of the plugin module as defined
        in :attr:`attr_unload` if available, otherwise a no-operation lambda will be returned.

        Returns:
            callable: Unload method for the plugin module.
        """
        return self._get_instance_attribute(
            self.__class__.attr_unload, default=lambda: True
        )

    @property
    def enable(self):
        """
        Method for enabling the plugin module. Will be taken from the enable attribute of the plugin module as defined
        in :attr:`attr_enable` if available, otherwise a no-operation lambda will be returned.

        Returns:
            callable: Enable method for the plugin module.
        """
        return self._get_instance_attribute(
            self.__class__.attr_enable, default=lambda: True
        )

    @property
    def disable(self):
        """
        Method for disabling the plugin module. Will be taken from the disable attribute of the plugin module as defined
        in :attr:`attr_disable` if available, otherwise a no-operation lambda will be returned.

        Returns:
            callable: Disable method for the plugin module.
        """
        return self._get_instance_attribute(
            self.__class__.attr_disable, default=lambda: True
        )

    def _get_instance_attribute(
        self, attr, default=None, defaults=None, incl_metadata=False
    ):
        if self.instance is None or not hasattr(self.instance, attr):
            if incl_metadata and attr in self.parsed_metadata:
                return self.parsed_metadata[attr]

            elif defaults is not None:
                for value in defaults:
                    if callable(value):
                        value = value()
                    if value is not None:
                        return value

            return default

        return getattr(self.instance, attr)

    @property
    def parsed_metadata(self):
        """The plugin metadata parsed from the plugin ``__init__.py``'s AST."""
        return self._cached_parsed_metadata

    def _parse_metadata(self):
        result = {}

        path = self.location
        if not path:
            return result

        if os.path.isdir(path):
            path = os.path.join(self.location, "__init__.py")

        if not os.path.isfile(path):
            return result

        if not path.endswith(".py"):
            # we only support parsing plain text source files
            return result

        self._logger.debug(
            "Parsing plugin metadata for {} from AST of {}".format(self.key, path)
        )

        try:
            import ast

            with io.open(path, "rb") as f:
                root = ast.parse(f.read(), filename=path)

            assignments = list(
                filter(lambda x: isinstance(x, ast.Assign) and x.targets, root.body)
            )

            def extract_target_ids(node):
                return list(
                    map(
                        lambda x: x.id,
                        filter(lambda x: isinstance(x, ast.Name), node.targets),
                    )
                )

            for key in (
                self.__class__.attr_name,
                self.__class__.attr_version,
                self.__class__.attr_author,
                self.__class__.attr_description,
                self.__class__.attr_url,
                self.__class__.attr_license,
                self.__class__.attr_pythoncompat,
            ):
                for a in reversed(assignments):
                    targets = extract_target_ids(a)
                    if key in targets:
                        if isinstance(a.value, ast.Str):
                            result[key] = a.value.s

                        elif (
                            isinstance(a.value, ast.Call)
                            and hasattr(a.value, "func")
                            and a.value.func.id == "gettext"
                            and a.value.args
                            and isinstance(a.value.args[0], ast.Str)
                        ):
                            result[key] = a.value.args[0].s

                        break

            for key in (self.__class__.attr_hidden,):
                for a in reversed(assignments):
                    targets = extract_target_ids(a)
                    if key in targets:
                        if isinstance(a.value, ast.Name) and a.value.id in (
                            "True",
                            "False",
                        ):
                            result[key] = bool(a.value.id)

                        break

        except SyntaxError:
            self._logger.exception(
                "Invalid syntax in {} for plugin {}".format(path, self.key)
            )
            self.invalid_syntax = True
        except Exception:
            self._logger.exception(
                "Error while parsing AST from {} for plugin {}".format(path, self.key)
            )

        return result


class PluginManager(object):
    """
    The :class:`PluginManager` is the central component for finding, loading and accessing plugins provided to the
    system.

    It is able to discover plugins both through possible file system locations as well as customizable entry points.
    """

    plugin_timings_logtarget = "PLUGIN_TIMINGS"
    plugin_timings_message = "{func} - {timing:05.2f}ms"

    def __init__(
        self,
        plugin_folders,
        plugin_bases,
        plugin_entry_points,
        logging_prefix=None,
        plugin_disabled_list=None,
        plugin_blacklist=None,
        plugin_restart_needing_hooks=None,
        plugin_obsolete_hooks=None,
        plugin_considered_bundled=None,
        plugin_validators=None,
        compatibility_ignored_list=None,
    ):
        self.logger = logging.getLogger(__name__)

        if logging_prefix is None:
            logging_prefix = ""
        if plugin_folders is None:
            plugin_folders = []
        if plugin_bases is None:
            plugin_bases = []
        if plugin_entry_points is None:
            plugin_entry_points = []
        if plugin_disabled_list is None:
            plugin_disabled_list = []
        if plugin_blacklist is None:
            plugin_blacklist = []
        if compatibility_ignored_list is None:
            compatibility_ignored_list = []
        if plugin_considered_bundled is None:
            plugin_considered_bundled = []

        self.plugin_folders = plugin_folders
        self.plugin_bases = plugin_bases
        self.plugin_entry_points = plugin_entry_points
        self.plugin_disabled_list = plugin_disabled_list
        self.plugin_blacklist = plugin_blacklist
        self.plugin_restart_needing_hooks = plugin_restart_needing_hooks
        self.plugin_obsolete_hooks = plugin_obsolete_hooks
        self.plugin_validators = plugin_validators
        self.logging_prefix = logging_prefix
        self.compatibility_ignored_list = compatibility_ignored_list
        self.plugin_considered_bundled = plugin_considered_bundled

        self.enabled_plugins = {}
        self.disabled_plugins = {}
        self.plugin_implementations = {}
        self.plugin_implementations_by_type = defaultdict(list)

        self._plugin_hooks = defaultdict(list)

        self.implementation_injects = {}
        self.implementation_inject_factories = []
        self.implementation_pre_inits = []
        self.implementation_post_inits = []

        self.on_plugin_loaded = lambda *args, **kwargs: None
        self.on_plugin_unloaded = lambda *args, **kwargs: None
        self.on_plugin_enabled = lambda *args, **kwargs: None
        self.on_plugin_disabled = lambda *args, **kwargs: None
        self.on_plugin_implementations_initialized = lambda *args, **kwargs: None

        self.on_plugins_loaded = lambda *args, **kwargs: None
        self.on_plugins_enabled = lambda *args, **kwargs: None

        self.registered_clients = []

        self.marked_plugins = defaultdict(list)

        self._python_install_dir = None
        self._python_virtual_env = False
        self._detect_python_environment()

    def _detect_python_environment(self):
        import sys
        from distutils.command.install import install as cmd_install
        from distutils.dist import Distribution

        cmd = cmd_install(Distribution())
        cmd.finalize_options()

        self._python_install_dir = cmd.install_lib
        self._python_prefix = os.path.realpath(sys.prefix)
        self._python_virtual_env = hasattr(sys, "real_prefix") or (
            hasattr(sys, "base_prefix")
            and os.path.realpath(sys.prefix) != os.path.realpath(sys.base_prefix)
        )

    @property
    def plugins(self):
        """
        Returns:
                (list) list of enabled and disabled registered plugins
        """
        plugins = dict(self.enabled_plugins)
        plugins.update(self.disabled_plugins)
        return plugins

    @property
    def plugin_hooks(self):
        """
        Returns:
                (dict) dictionary of registered hooks and their handlers
        """
        return {
            key: list(map(lambda v: (v[1], v[2]), value))
            for key, value in self._plugin_hooks.items()
        }

    def find_plugins(self, existing=None, ignore_uninstalled=True, incl_all_found=False):
        added, found = self._find_plugins(
            existing=existing, ignore_uninstalled=ignore_uninstalled
        )
        if incl_all_found:
            return added, found
        else:
            return added

    def _find_plugins(self, existing=None, ignore_uninstalled=True):
        if existing is None:
            existing = dict(self.plugins)

        result_added = OrderedDict()
        result_found = []

        if self.plugin_folders:
            try:
                added, found = self._find_plugins_from_folders(
                    self.plugin_folders, existing, ignored_uninstalled=ignore_uninstalled
                )
                result_added.update(added)
                result_found += found
            except Exception:
                self.logger.exception("Error fetching plugins from folders")

        if self.plugin_entry_points:
            existing.update(result_added)
            try:
                added, found = self._find_plugins_from_entry_points(
                    self.plugin_entry_points,
                    existing,
                    ignore_uninstalled=ignore_uninstalled,
                )
                result_added.update(added)
                result_found += found
            except Exception:
                self.logger.exception("Error fetching plugins from entry points")

        return result_added, result_found

    def _find_plugins_from_folders(self, folders, existing, ignored_uninstalled=True):
        added = OrderedDict()
        found = []

        for folder in folders:
            try:
                flagged_readonly = False
                package = None
                if isinstance(folder, (list, tuple)):
                    if len(folder) == 2:
                        folder, flagged_readonly = folder
                    elif len(folder) == 3:
                        folder, package, flagged_readonly = folder
                    else:
                        continue
                actual_readonly = not os.access(folder, os.W_OK)

                if not os.path.exists(folder):
                    self.logger.warning(
                        "Plugin folder {folder} could not be found, skipping it".format(
                            folder=folder
                        )
                    )
                    continue

                for entry in scandir(folder):
                    try:
                        if entry.is_dir():
                            init_py = os.path.join(entry.path, "__init__.py")
                            init_pyc = os.path.join(entry.path, "__init__.pyc")

                            if not os.path.isfile(init_py) and not os.path.isfile(
                                init_pyc
                            ):
                                # neither does exist, we ignore this
                                continue

                            key = entry.name

                        elif entry.is_file():
                            key, ext = os.path.splitext(entry.name)
                            if ext not in (".py", ".pyc") or key.startswith("__"):
                                # neither py nor pyc, or starts with __ (like __init__), we ignore this
                                continue

                        else:
                            # whatever this is, we ignore it
                            continue

                        found.append(key)
                        if (
                            key in existing
                            or key in added
                            or (
                                ignored_uninstalled
                                and key in self.marked_plugins["uninstalled"]
                            )
                        ):
                            # plugin is already defined, ignore it
                            continue

                        bundled = flagged_readonly

                        module_name = None
                        if package:
                            module_name = "{}.{}".format(package, key)

                        plugin = self._import_plugin_from_module(
                            key, module_name=module_name, folder=folder, bundled=bundled
                        )
                        if plugin:
                            if module_name:
                                plugin.origin = ModuleOrigin(
                                    "module", module_name, folder
                                )
                            else:
                                plugin.origin = FolderOrigin("folder", folder)
                            plugin.managable = (
                                not flagged_readonly and not actual_readonly
                            )
                            plugin.enabled = False
                            added[key] = plugin
                    except Exception:
                        self.logger.exception(
                            "Error processing folder entry {!r} from folder {}".format(
                                entry, folder
                            )
                        )
            except Exception:
                self.logger.exception("Error processing folder {}".format(folder))

        return added, found

    def _find_plugins_from_entry_points(self, groups, existing, ignore_uninstalled=True):
        added = OrderedDict()
        found = []

        # let's make sure we have a current working set ...
        working_set = pkg_resources.WorkingSet()

        # ... including the user's site packages
        import site
        import sys

        if site.ENABLE_USER_SITE:
            if site.USER_SITE not in working_set.entries:
                working_set.add_entry(site.USER_SITE)
            if site.USER_SITE not in sys.path:
                site.addsitedir(site.USER_SITE)

        if not isinstance(groups, (list, tuple)):
            groups = [groups]

        def wrapped(gen):
            # to protect against some issues in installed packages that make iteration over entry points
            # fall on its face - e.g. https://groups.google.com/forum/#!msg/octoprint/DyXdqhR0U7c/kKMUsMmIBgAJ
            for entry in gen:
                try:
                    yield entry
                except Exception:
                    self.logger.exception(
                        "Something went wrong while processing the entry points of a package in the "
                        "Python environment - broken entry_points.txt in some package?"
                    )

        for group in groups:
            for entry_point in wrapped(
                working_set.iter_entry_points(group=group, name=None)
            ):
                try:
                    key = entry_point.name
                    module_name = entry_point.module_name
                    version = entry_point.dist.version

                    found.append(key)
                    if (
                        key in existing
                        or key in added
                        or (
                            ignore_uninstalled
                            and key in self.marked_plugins["uninstalled"]
                        )
                    ):
                        # plugin is already defined or marked as uninstalled, ignore it
                        continue

                    bundled = key in self.plugin_considered_bundled
                    kwargs = {
                        "module_name": module_name,
                        "version": version,
                        "bundled": bundled,
                    }
                    package_name = entry_point.dist.project_name
                    try:
                        entry_point_metadata = EntryPointMetadata(entry_point)
                    except Exception:
                        self.logger.exception(
                            "Something went wrong while retrieving metadata for module {}".format(
                                module_name
                            )
                        )
                    else:
                        kwargs.update(
                            {
                                "name": entry_point_metadata.name,
                                "summary": entry_point_metadata.summary,
                                "author": entry_point_metadata.author,
                                "url": entry_point_metadata.home_page,
                                "license": entry_point_metadata.license,
                            }
                        )

                    plugin = self._import_plugin_from_module(key, **kwargs)
                    if plugin:
                        plugin.origin = EntryPointOrigin(
                            "entry_point", group, module_name, package_name, version
                        )
                        plugin.enabled = False

                        # plugin is manageable if its location is writable and OctoPrint
                        # is either not running from a virtual env or the plugin is
                        # installed in that virtual env - the virtual env's pip will not
                        # allow us to uninstall stuff that is installed outside
                        # of the virtual env, so this check is necessary
                        plugin.managable = os.access(plugin.location, os.W_OK) and (
                            not self._python_virtual_env
                            or is_sub_path_of(plugin.location, self._python_prefix)
                            or is_editable_install(
                                self._python_install_dir,
                                package_name,
                                module_name,
                                plugin.location,
                            )
                        )

                        added[key] = plugin
                except Exception:
                    self.logger.exception(
                        "Error processing entry point {!r} for group {}".format(
                            entry_point, group
                        )
                    )

        return added, found

    def _import_plugin_from_module(
        self,
        key,
        folder=None,
        module_name=None,
        name=None,
        version=None,
        summary=None,
        author=None,
        url=None,
        license=None,
        bundled=False,
    ):
        # TODO error handling
        try:
            if folder:
                location, spec = _find_module(key, path=folder)
            elif module_name:
                location, spec = _find_module(module_name)
            else:
                return None
        except Exception:
            self.logger.exception("Could not locate plugin {key}".format(key=key))
            return None

        # Create a simple dummy entry first ...
        plugin = PluginInfo(
            key,
            location,
            None,
            name=name,
            version=version,
            description=summary,
            author=author,
            url=url,
            license=license,
        )
        plugin.bundled = bundled

        if self._is_plugin_disabled(key):
            self.logger.info("Plugin {} is disabled.".format(plugin))
            plugin.forced_disabled = True

        if self._is_plugin_blacklisted(key) or (
            plugin.version is not None
            and self._is_plugin_version_blacklisted(key, plugin.version)
        ):
            self.logger.warning("Plugin {} is blacklisted.".format(plugin))
            plugin.blacklisted = True

        python_version = get_python_version_string()
        if self._is_plugin_incompatible(key, plugin):
            if plugin.invalid_syntax:
                self.logger.warning(
                    "Plugin {} can't be compiled under Python {} due to invalid syntax".format(
                        plugin, python_version
                    )
                )
            else:
                self.logger.warning(
                    "Plugin {} is not compatible to Python {} (compatibility string: {}).".format(
                        plugin, python_version, plugin.pythoncompat
                    )
                )
            plugin.incompatible = True

        if not plugin.validate(
            "before_import", additional_validators=self.plugin_validators
        ):
            return plugin

        # ... then create and return the real one
        return self._import_plugin(
            key,
            spec,
            module_name=module_name,
            name=name,
            location=plugin.location,
            version=version,
            summary=summary,
            author=author,
            url=url,
            license=license,
            bundled=bundled,
            parsed_metadata=plugin.parsed_metadata,
        )

    def _import_plugin(
        self,
        key,
        spec,
        module_name=None,
        name=None,
        location=None,
        version=None,
        summary=None,
        author=None,
        url=None,
        license=None,
        bundled=False,
        parsed_metadata=None,
    ):
        try:
            if module_name:
                module = _load_module(module_name, spec)
            else:
                module = _load_module(key, spec)

            plugin = PluginInfo(
                key,
                location,
                module,
                name=name,
                version=version,
                description=summary,
                author=author,
                url=url,
                license=license,
                parsed_metadata=parsed_metadata,
            )

            plugin.bundled = bundled
        except Exception:
            self.logger.exception("Error loading plugin {key}".format(key=key))
            return None

        if plugin.check():
            return plugin
        else:
            self.logger.info(
                "Plugin {plugin} did not pass check, not loading.".format(
                    plugin=str(plugin)
                )
            )
            return None

    def _is_plugin_disabled(self, key):
        return key in self.plugin_disabled_list or key.endswith("disabled")

    def _is_plugin_blacklisted(self, key):
        return key in self.plugin_blacklist

    def _is_plugin_incompatible(self, key, plugin):
        return (
            not plugin.bundled
            and not is_python_compatible(plugin.pythoncompat)
            and key not in self.compatibility_ignored_list
        )

    def _is_plugin_version_blacklisted(self, key, version):
        def matches_plugin(entry):
            if isinstance(entry, (tuple, list)) and len(entry) == 2:
                entry_key, entry_version = entry
                return entry_key == key and entry_version == version
            return False

        return any(map(lambda entry: matches_plugin(entry), self.plugin_blacklist))

    def reload_plugins(
        self, startup=False, initialize_implementations=True, force_reload=None
    ):
        """
        Reloads plugins, detecting newly added ones in the process.

        Args:
                startup (boolean): whether this is called during startup of the platform
                initialize_implementations (boolean): whether plugin implementations should be initialized
                force_reload (list): list of plugin identifiers which should be force reloaded
        """
        self.logger.info(
            "Loading plugins from {folders} and installed plugin packages...".format(
                folders=", ".join(
                    map(
                        lambda x: x[0] if isinstance(x, tuple) else str(x),
                        self.plugin_folders,
                    )
                )
            )
        )

        if force_reload is None:
            force_reload = []

        added, found = self.find_plugins(
            existing=dict(
                (k, v) for k, v in self.plugins.items() if k not in force_reload
            ),
            incl_all_found=True,
        )

        # let's clean everything we DIDN'T find first
        removed = [
            key
            for key in list(self.enabled_plugins.keys())
            + list(self.disabled_plugins.keys())
            if key not in found
        ]
        for key in removed:
            try:
                del self.enabled_plugins[key]
            except KeyError:
                pass

            try:
                del self.disabled_plugins[key]
            except KeyError:
                pass

        self.disabled_plugins.update(added)

        # 1st pass: loading the plugins
        for name, plugin in added.items():
            try:
                if (
                    not plugin.blacklisted
                    and not plugin.forced_disabled
                    and not plugin.incompatible
                ):
                    self.load_plugin(
                        name,
                        plugin,
                        startup=startup,
                        initialize_implementation=initialize_implementations,
                    )
            except PluginNeedsRestart:
                pass
            except PluginLifecycleException as e:
                self.logger.info(str(e))

        self.on_plugins_loaded(
            startup=startup,
            initialize_implementations=initialize_implementations,
            force_reload=force_reload,
        )

        # 2nd pass: enabling those plugins that need enabling
        for name, plugin in added.items():
            try:
                if (
                    plugin.loaded
                    and not plugin.forced_disabled
                    and not plugin.incompatible
                ):
                    if plugin.blacklisted:
                        self.logger.warning(
                            "Plugin {} is blacklisted. Not enabling it.".format(plugin)
                        )
                        continue
                    self.enable_plugin(
                        name,
                        plugin=plugin,
                        initialize_implementation=initialize_implementations,
                        startup=startup,
                    )
            except PluginNeedsRestart:
                pass
            except PluginLifecycleException as e:
                self.logger.info(str(e))

        self.on_plugins_enabled(
            startup=startup,
            initialize_implementations=initialize_implementations,
            force_reload=force_reload,
        )

        if len(self.enabled_plugins) <= 0:
            self.logger.info("No plugins found")
        else:
            self.logger.info(
                "Found {count} plugin(s) providing {implementations} mixin implementations, {hooks} hook handlers".format(
                    count=len(self.enabled_plugins) + len(self.disabled_plugins),
                    implementations=len(self.plugin_implementations),
                    hooks=sum(map(lambda x: len(x), self.plugin_hooks.values())),
                )
            )

    def mark_plugin(self, name, **flags):
        """
        Mark plugin ``name`` with an arbitrary number of flags.

        Args:
                name (str): plugin identifier
                **flags (dict): dictionary of flag names and values
        """
        if name not in self.plugins:
            self.logger.debug(
                "Trying to mark an unknown plugin {name}".format(**locals())
            )

        for key, value in flags.items():
            if value is None:
                continue

            if value and name not in self.marked_plugins[key]:
                self.marked_plugins[key].append(name)
            elif not value and name in self.marked_plugins[key]:
                self.marked_plugins[key].remove(name)

    def is_plugin_marked(self, name, flag):
        """
        Checks whether a plugin has been marked with a certain flag.

        Args:
                name (str): the plugin's identifier
                flag (str): the flag to check

        Returns:
                (boolean): True if the plugin has been flagged, False otherwise
        """
        if name not in self.plugins:
            return False

        return name in self.marked_plugins[flag]

    def load_plugin(
        self, name, plugin=None, startup=False, initialize_implementation=True
    ):
        if name not in self.plugins:
            self.logger.warning(
                "Trying to load an unknown plugin {name}".format(**locals())
            )
            return

        if plugin is None:
            plugin = self.plugins[name]

        try:
            if not plugin.validate(
                "before_load", additional_validators=self.plugin_validators
            ):
                return

            plugin.load()
            plugin.validate("after_load", additional_validators=self.plugin_validators)
            self.on_plugin_loaded(name, plugin)
            plugin.loaded = True

            self.logger.debug("Loaded plugin {name}: {plugin}".format(**locals()))
        except PluginLifecycleException as e:
            raise e
        except Exception:
            self.logger.exception("There was an error loading plugin %s" % name)

    def unload_plugin(self, name):
        if name not in self.plugins:
            self.logger.warning(
                "Trying to unload unknown plugin {name}".format(**locals())
            )
            return

        plugin = self.plugins[name]

        try:
            if plugin.enabled:
                self.disable_plugin(name, plugin=plugin)

            plugin.unload()
            self.on_plugin_unloaded(name, plugin)

            if name in self.enabled_plugins:
                del self.enabled_plugins[name]

            if name in self.disabled_plugins:
                del self.disabled_plugins[name]

            plugin.loaded = False

            self.logger.debug("Unloaded plugin {name}: {plugin}".format(**locals()))
        except PluginLifecycleException as e:
            raise e
        except Exception:
            self.logger.exception(
                "There was an error unloading plugin {name}".format(**locals())
            )

            # make sure the plugin is NOT in the list of enabled plugins but in the list of disabled plugins
            if name in self.enabled_plugins:
                del self.enabled_plugins[name]
            if name not in self.disabled_plugins:
                self.disabled_plugins[name] = plugin

    def enable_plugin(
        self, name, plugin=None, initialize_implementation=True, startup=False
    ):
        """Enables a plugin"""
        if name not in self.disabled_plugins:
            self.logger.warning(
                "Tried to enable plugin {name}, however it is not disabled".format(
                    **locals()
                )
            )
            return

        if plugin is None:
            plugin = self.disabled_plugins[name]

        if not startup and self.is_restart_needing_plugin(plugin):
            raise PluginNeedsRestart(name)

        if self.has_obsolete_hooks(plugin):
            raise PluginCantEnable(
                name,
                "Dependency on obsolete hooks detected, full functionality cannot be guaranteed",
            )

        try:
            if not plugin.validate(
                "before_enable", additional_validators=self.plugin_validators
            ):
                return False

            plugin.enable()
            self._activate_plugin(name, plugin)
        except PluginLifecycleException as e:
            raise e
        except Exception:
            self.logger.exception(
                "There was an error while enabling plugin {name}".format(**locals())
            )
            return False
        else:
            if name in self.disabled_plugins:
                del self.disabled_plugins[name]
            self.enabled_plugins[name] = plugin
            plugin.enabled = True

            if plugin.implementation:
                if initialize_implementation:
                    if not self.initialize_implementation_of_plugin(name, plugin):
                        return False
                plugin.implementation.on_plugin_enabled()
            self.on_plugin_enabled(name, plugin)

            self.logger.debug("Enabled plugin {name}: {plugin}".format(**locals()))

        return True

    def disable_plugin(self, name, plugin=None):
        """Disables a plugin"""
        if name not in self.enabled_plugins:
            self.logger.warning(
                "Tried to disable plugin {name}, however it is not enabled".format(
                    **locals()
                )
            )
            return

        if plugin is None:
            plugin = self.enabled_plugins[name]

        if self.is_restart_needing_plugin(plugin):
            raise PluginNeedsRestart(name)

        try:
            plugin.disable()
            self._deactivate_plugin(name, plugin)
        except PluginLifecycleException as e:
            raise e
        except Exception:
            self.logger.exception(
                "There was an error while disabling plugin {name}".format(**locals())
            )
            return False
        else:
            if name in self.enabled_plugins:
                del self.enabled_plugins[name]
            self.disabled_plugins[name] = plugin
            plugin.enabled = False

            if plugin.implementation:
                plugin.implementation.on_plugin_disabled()
            self.on_plugin_disabled(name, plugin)

            self.logger.debug("Disabled plugin {name}: {plugin}".format(**locals()))

        return True

    def _activate_plugin(self, name, plugin):
        plugin.hotchangeable = self.is_restart_needing_plugin(plugin)

        # evaluate registered hooks
        for hook, definition in plugin.hooks.items():
            try:
                callback, order = self._get_callback_and_order(definition)
            except ValueError as e:
                self.logger.warning(
                    "There is something wrong with the hook definition {} for plugin {}: {}".format(
                        definition, name, str(e)
                    )
                )
                continue

            self._plugin_hooks[hook].append((order, name, callback))
            self._sort_hooks(hook)

        # evaluate registered implementation
        if plugin.implementation:
            mixins = self.mixins_matching_bases(
                plugin.implementation.__class__, *self.plugin_bases
            )
            for mixin in mixins:
                self.plugin_implementations_by_type[mixin].append(
                    (name, plugin.implementation)
                )
                if not getattr(plugin.implementation, "__timing_wrapped", False):
                    for method in filter(
                        lambda a: not a.startswith("_") and callable(getattr(mixin, a)),
                        dir(mixin),
                    ):
                        # wrap method with time_this
                        setattr(
                            plugin.implementation,
                            method,
                            time_this(
                                logtarget=self.plugin_timings_logtarget,
                                expand_logtarget=True,
                                message=self.plugin_timings_message,
                            )(getattr(plugin.implementation, method)),
                        )

            self.plugin_implementations[name] = plugin.implementation
            plugin.implementation.__timing_wrapped = True

    def _deactivate_plugin(self, name, plugin):
        for hook, definition in plugin.hooks.items():
            try:
                callback, order = self._get_callback_and_order(definition)
            except ValueError as e:
                self.logger.warning(
                    "There is something wrong with the hook definition {} for plugin {}: {}".format(
                        definition, name, str(e)
                    )
                )
                continue

            try:
                self._plugin_hooks[hook].remove((order, name, callback))
                self._sort_hooks(hook)
            except ValueError:
                # that's ok, the plugin was just not registered for the hook
                pass

        if plugin.implementation is not None:
            if name in self.plugin_implementations:
                del self.plugin_implementations[name]

            mixins = self.mixins_matching_bases(
                plugin.implementation.__class__, *self.plugin_bases
            )
            for mixin in mixins:
                try:
                    self.plugin_implementations_by_type[mixin].remove(
                        (name, plugin.implementation)
                    )
                except ValueError:
                    # that's ok, the plugin was just not registered for the type
                    pass

    def is_restart_needing_plugin(self, plugin):
        """Checks whether the plugin needs a restart on changes"""
        return (
            plugin.needs_restart
            or self.has_restart_needing_implementation(plugin)
            or self.has_restart_needing_hooks(plugin)
        )

    def has_restart_needing_implementation(self, plugin):
        """Checks whether the plugin's implementation needs a restart on changes"""
        return self.has_any_of_mixins(plugin, RestartNeedingPlugin)

    def has_restart_needing_hooks(self, plugin):
        """Checks whether the plugin has any hooks that need a restart on changes"""
        return self.has_any_of_hooks(plugin, self.plugin_restart_needing_hooks)

    def has_obsolete_hooks(self, plugin):
        """Checks whether the plugin uses any obsolete hooks"""
        return self.has_any_of_hooks(plugin, self.plugin_obsolete_hooks)

    def is_restart_needing_hook(self, hook):
        """Checks whether a hook needs a restart on changes"""
        return self.hook_matches_hooks(hook, self.plugin_restart_needing_hooks)

    def is_obsolete_hook(self, hook):
        """Checks whether a hook is obsolete"""
        return self.hook_matches_hooks(hook, self.plugin_obsolete_hooks)

    @staticmethod
    def has_any_of_hooks(plugin, *hooks):
        """
        Tests if the ``plugin`` contains any of the provided ``hooks``.

        Uses :func:`octoprint.plugin.core.PluginManager.hook_matches_hooks`.

        Args:
                plugin: plugin to test hooks for
                *hooks: hooks to test against

        Returns:
                (bool): True if any of the plugin's hooks match the provided hooks,
                        False otherwise.
        """

        if hooks and len(hooks) == 1 and isinstance(hooks[0], (list, tuple)):
            hooks = hooks[0]

        hooks = list(filter(lambda hook: hook is not None, hooks))
        if not hooks:
            return False
        if not plugin or not plugin.hooks:
            return False

        plugin_hooks = plugin.hooks.keys()

        return any(
            map(lambda hook: PluginManager.hook_matches_hooks(hook, *hooks), plugin_hooks)
        )

    @staticmethod
    def hook_matches_hooks(hook, *hooks):
        """
        Tests if ``hook`` matches any of the provided ``hooks`` to test for.

        ``hook`` is expected to be an exact hook name.

        ``hooks`` is expected to be a list containing one or more hook names or
        patterns. That can be either an exact hook name or an
        :func:`fnmatch.fnmatch` pattern.

        Args:
                hook: the hook to test
                hooks: the hook name patterns to test against

        Returns:
                (bool): True if the ``hook`` matches any of the ``hooks``, False otherwise.

        """

        if hooks and len(hooks) == 1 and isinstance(hooks[0], (list, tuple)):
            hooks = hooks[0]

        hooks = list(filter(lambda hook: hook is not None, hooks))
        if not hooks:
            return False
        if not hook:
            return False

        return any(map(lambda h: fnmatch.fnmatch(hook, h), hooks))

    @staticmethod
    def mixins_matching_bases(klass, *bases):
        result = set()
        for c in inspect.getmro(klass):
            if c == klass or c in bases:
                # ignore the exact class and our bases
                continue
            if issubclass(c, bases):
                result.add(c)
        return result

    @staticmethod
    def has_any_of_mixins(plugin, *mixins):
        """
        Tests if the ``plugin`` has an implementation implementing any
        of the provided ``mixins``.

        Args:
                plugin: plugin for which to check the implementation
                *mixins: mixins to test against

        Returns:
                (bool): True if the plugin's implementation implements any of the
                        provided mixins, False otherwise.
        """

        if mixins and len(mixins) == 1 and isinstance(mixins[0], (list, tuple)):
            mixins = mixins[0]

        mixins = list(filter(lambda mixin: mixin is not None, mixins))
        if not mixins:
            return False
        if not plugin or not plugin.implementation:
            return False

        return isinstance(plugin.implementation, tuple(mixins))

    def initialize_implementations(
        self,
        additional_injects=None,
        additional_inject_factories=None,
        additional_pre_inits=None,
        additional_post_inits=None,
    ):
        for name, plugin in self.enabled_plugins.items():
            self.initialize_implementation_of_plugin(
                name,
                plugin,
                additional_injects=additional_injects,
                additional_inject_factories=additional_inject_factories,
                additional_pre_inits=additional_pre_inits,
                additional_post_inits=additional_post_inits,
            )

        self.logger.info(
            "Initialized {count} plugin implementation(s)".format(
                count=len(self.plugin_implementations)
            )
        )

    def initialize_implementation_of_plugin(
        self,
        name,
        plugin,
        additional_injects=None,
        additional_inject_factories=None,
        additional_pre_inits=None,
        additional_post_inits=None,
    ):
        if plugin.implementation is None:
            return

        return self.initialize_implementation(
            name,
            plugin,
            plugin.implementation,
            additional_injects=additional_injects,
            additional_inject_factories=additional_inject_factories,
            additional_pre_inits=additional_pre_inits,
            additional_post_inits=additional_post_inits,
        )

    def initialize_implementation(
        self,
        name,
        plugin,
        implementation,
        additional_injects=None,
        additional_inject_factories=None,
        additional_pre_inits=None,
        additional_post_inits=None,
    ):
        if additional_injects is None:
            additional_injects = {}
        if additional_inject_factories is None:
            additional_inject_factories = []
        if additional_pre_inits is None:
            additional_pre_inits = []
        if additional_post_inits is None:
            additional_post_inits = []

        injects = self.implementation_injects
        injects.update(additional_injects)

        inject_factories = self.implementation_inject_factories
        inject_factories += additional_inject_factories

        pre_inits = self.implementation_pre_inits
        pre_inits += additional_pre_inits

        post_inits = self.implementation_post_inits
        post_inits += additional_post_inits

        try:
            kwargs = dict(injects)

            kwargs.update(
                {
                    "identifier": name,
                    "plugin_name": plugin.name,
                    "plugin_version": plugin.version,
                    "plugin_info": plugin,
                    "basefolder": os.path.realpath(plugin.location),
                    "logger": logging.getLogger(self.logging_prefix + name),
                }
            )

            # inject the additional_injects
            for arg, value in kwargs.items():
                setattr(implementation, "_" + arg, value)

            # inject any injects produced in the additional_inject_factories
            for factory in inject_factories:
                try:
                    return_value = factory(name, implementation)
                except Exception:
                    self.logger.exception(
                        "Exception while executing injection factory %r" % factory
                    )
                else:
                    if return_value is not None:
                        if isinstance(return_value, dict):
                            for arg, value in return_value.items():
                                setattr(implementation, "_" + arg, value)

            # execute any additional pre init methods
            for pre_init in pre_inits:
                pre_init(name, implementation)

            implementation.initialize()

            # execute any additional post init methods
            for post_init in post_inits:
                post_init(name, implementation)

        except Exception as e:
            self._deactivate_plugin(name, plugin)
            plugin.enabled = False

            if isinstance(e, PluginLifecycleException):
                raise e
            else:
                self.logger.exception(
                    "Exception while initializing plugin {name}, disabling it".format(
                        **locals()
                    )
                )
                return False
        else:
            self.on_plugin_implementations_initialized(name, plugin)

        self.logger.debug(
            "Initialized plugin mixin implementation for plugin {name}".format(**locals())
        )
        return True

    def log_all_plugins(
        self,
        show_bundled=True,
        bundled_str=(" (bundled)", ""),
        show_location=True,
        location_str=" = {location}",
        show_enabled=True,
        enabled_str=(" ", "!", "#", "*"),
        only_to_handler=None,
    ):
        all_plugins = list(self.enabled_plugins.values()) + list(
            self.disabled_plugins.values()
        )

        def _log(message, level=logging.INFO):
            if only_to_handler is not None:
                import octoprint.logging

                octoprint.logging.log_to_handler(
                    self.logger, only_to_handler, level, message, []
                )
            else:
                self.logger.log(level, message)

        if len(all_plugins) <= 0:
            _log("No plugins available")
        else:
            formatted_plugins = "\n".join(
                map(
                    lambda x: "| "
                    + x.long_str(
                        show_bundled=show_bundled,
                        bundled_strs=bundled_str,
                        show_location=show_location,
                        location_str=location_str,
                        show_enabled=show_enabled,
                        enabled_strs=enabled_str,
                    ),
                    sorted(self.plugins.values(), key=lambda x: unicode(x).lower()),
                )
            )
            legend = "Prefix legend: {1} = disabled, {2} = blacklisted, {3} = incompatible".format(
                *enabled_str
            )
            _log(
                "{count} plugin(s) registered with the system:\n{plugins}\n{legend}".format(
                    count=len(all_plugins), plugins=formatted_plugins, legend=legend
                )
            )

    def get_plugin(self, identifier, require_enabled=True):
        """
        Retrieves the module of the plugin identified by ``identifier``. If the plugin is not registered or disabled and
        ``required_enabled`` is True (the default) None will be returned.

        Arguments:
            identifier (str): The identifier of the plugin to retrieve.
            require_enabled (boolean): Whether to only return the plugin if is enabled (True, default) or also if it's
                disabled.

        Returns:
            module: The requested plugin module or None
        """

        plugin_info = self.get_plugin_info(identifier, require_enabled=require_enabled)
        if plugin_info is not None:
            return plugin_info.instance
        return None

    def get_plugin_info(self, identifier, require_enabled=True):
        """
        Retrieves the :class:`PluginInfo` instance identified by ``identifier``. If the plugin is not registered or
        disabled and ``required_enabled`` is True (the default) None will be returned.

        Arguments:
            identifier (str): The identifier of the plugin to retrieve.
            require_enabled (boolean): Whether to only return the plugin if is enabled (True, default) or also if it's
                disabled.

        Returns:
            ~.PluginInfo: The requested :class:`PluginInfo` or None
        """

        if identifier in self.enabled_plugins:
            return self.enabled_plugins[identifier]
        elif not require_enabled and identifier in self.disabled_plugins:
            return self.disabled_plugins[identifier]

        return None

    def get_hooks(self, hook):
        """
        Retrieves all registered handlers for the specified hook.

        Arguments:
            hook (str): The hook for which to retrieve the handlers.

        Returns:
            dict: A dict containing all registered handlers mapped by their plugin's identifier.
        """

        if hook not in self.plugin_hooks:
            return {}

        result = OrderedDict()
        for h in self.plugin_hooks[hook]:
            result[h[0]] = time_this(
                logtarget=self.plugin_timings_logtarget,
                expand_logtarget=True,
                message=self.plugin_timings_message,
            )(h[1])
        return result

    def get_implementations(self, *types, **kwargs):
        """
        Get all mixin implementations that implement *all* of the provided ``types``.

        Arguments:
            types (one or more type): The types a mixin implementation needs to implement in order to be returned.

        Returns:
            list: A list of all found implementations
        """

        sorting_context = kwargs.get("sorting_context", None)

        result = None

        for t in types:
            implementations = self.plugin_implementations_by_type[t]
            if result is None:
                result = set(implementations)
            else:
                result = result.intersection(implementations)

        if result is None:
            return []

        def sort_func(impl):
            sorting_value = None
            if sorting_context is not None and isinstance(impl[1], SortablePlugin):
                try:
                    sorting_value = impl[1].get_sorting_key(sorting_context)
                except Exception:
                    self.logger.exception(
                        "Error while trying to retrieve sorting order for plugin {}".format(
                            impl[0]
                        )
                    )

                if sorting_value is not None:
                    try:
                        sorting_value = int(sorting_value)
                    except ValueError:
                        self.logger.warning(
                            "The order value returned by {} for sorting context {} is not a valid integer, ignoring it".format(
                                impl[0], sorting_context
                            )
                        )
                        sorting_value = None

            plugin_info = self.get_plugin_info(impl[0], require_enabled=False)
            return (
                sorting_value is None,
                sv(sorting_value),
                not plugin_info.bundled if plugin_info else True,
                sv(impl[0]),
            )

        return [impl[1] for impl in sorted(result, key=sort_func)]

    def get_filtered_implementations(self, f, *types, **kwargs):
        """
        Get all mixin implementations that implement *all* of the provided ``types`` and match the provided filter `f`.

        Arguments:
            f (callable): A filter function returning True for implementations to return and False for those to exclude.
            types (one or more type): The types a mixin implementation needs to implement in order to be returned.

        Returns:
            list: A list of all found and matching implementations.
        """

        assert callable(f)
        implementations = self.get_implementations(
            *types, sorting_context=kwargs.get("sorting_context", None)
        )
        return list(filter(f, implementations))

    def get_helpers(self, name, *helpers):
        """
        Retrieves the named ``helpers`` for the plugin with identifier ``name``.

        If the plugin is not available, returns None. Otherwise returns a :class:`dict` with the requested plugin
        helper names mapped to the method - if a helper could not be resolved, it will be missing from the dict.

        Arguments:
            name (str): Identifier of the plugin for which to look up the ``helpers``.
            helpers (one or more str): Identifiers of the helpers of plugin ``name`` to return.

        Returns:
            dict: A dictionary of all resolved helpers, mapped by their identifiers, or None if the plugin was not
                registered with the system.
        """

        if name not in self.enabled_plugins:
            return None
        plugin = self.enabled_plugins[name]

        all_helpers = plugin.helpers
        if len(helpers):
            return dict((k, v) for (k, v) in all_helpers.items() if k in helpers)
        else:
            return all_helpers

    def register_message_receiver(self, client):
        """
        Registers a ``client`` for receiving plugin messages. The ``client`` needs to be a callable accepting two
        input arguments, ``plugin`` (the sending plugin's identifier) and ``data`` (the message itself), and one
        optional keyword argument, ``permissions`` (an optional list of permissions to test against).
        """

        if client is None:
            return
        self.registered_clients.append(client)

    def unregister_message_receiver(self, client):
        """
        Unregisters a ``client`` for receiving plugin messages.
        """

        try:
            self.registered_clients.remove(client)
        except ValueError:
            # not registered
            pass

    def send_plugin_message(self, plugin, data, permissions=None):
        """
        Sends ``data`` in the name of ``plugin`` to all currently registered message receivers by invoking them
        with the three arguments.

        Arguments:
            plugin (str): The sending plugin's identifier.
            data (object): The message.
            permissions (list): A list of permissions to test against in the client.
        """

        for client in self.registered_clients:
            try:
                client(plugin, data, permissions=permissions)
            except Exception:
                self.logger.exception("Exception while sending plugin data to client")

    def _sort_hooks(self, hook):
        self._plugin_hooks[hook] = sorted(
            self._plugin_hooks[hook],
            key=lambda x: (x[0] is None, sv(x[0]), sv(x[1]), sv(x[2])),
        )

    def _get_callback_and_order(self, hook):
        if callable(hook):
            return hook, None

        elif isinstance(hook, tuple) and len(hook) == 2:
            callback, order = hook

            # test that callback is a callable
            if not callable(callback):
                raise ValueError("Hook callback is not a callable")

            # test that number is an int
            try:
                int(order)
            except ValueError:
                raise ValueError("Hook order is not a number")

            return callback, order

        else:
            raise ValueError(
                "Invalid hook definition, neither a callable nor a 2-tuple (callback, order): {!r}".format(
                    hook
                )
            )


def is_sub_path_of(path, parent):
    """
    Tests if `path` is a sub path (or identical) to `path`.

    >>> is_sub_path_of("/a/b/c", "/a/b")
    True
    >>> is_sub_path_of("/a/b/c", "/a/b2")
    False
    >>> is_sub_path_of("/a/b/c", "/b/c")
    False
    >>> is_sub_path_of("/foo/bar/../../a/b/c", "/a/b")
    True
    >>> is_sub_path_of("/a/b", "/a/b")
    True
    """
    rel_path = os.path.relpath(os.path.realpath(path), os.path.realpath(parent))
    return not (rel_path == os.pardir or rel_path.startswith(os.pardir + os.sep))


def is_editable_install(install_dir, package, module, location):
    package_link = os.path.join(install_dir, "{}.egg-link".format(package))
    if os.path.isfile(package_link):
        expected_target = os.path.normcase(os.path.realpath(location))
        try:
            with io.open(package_link, "rt", encoding="utf-8") as f:
                contents = f.readlines()
            for line in contents:
                target = os.path.normcase(
                    os.path.realpath(os.path.join(line.strip(), module))
                )
                if target == expected_target:
                    return True
        except Exception:
            raise  # TODO really ignore this?
            pass
    return False


class EntryPointMetadata(pkginfo.Distribution):
    def __init__(self, entry_point):
        self.entry_point = entry_point
        self.extractMetadata()

    def read(self):
        import warnings

        metadata_files = ("METADATA", "PKG-INFO")  # wheel  # egg

        if self.entry_point and self.entry_point.dist:
            for metadata_file in metadata_files:
                try:
                    return self.entry_point.dist.get_metadata(metadata_file)
                except (IOError, OSError):  # noqa: B014
                    # file not found, metadata file might be missing, ignore
                    # IOError: file not found in Py2
                    # OSError (specifically FileNotFoundError): file not found in Py3
                    pass

        warnings.warn(
            "No package metadata found for package {}".format(
                self.entry_point.module_name
            )
        )


class Plugin(object):
    """
    The parent class of all plugin implementations.

    .. attribute:: _identifier

       The identifier of the plugin. Injected by the plugin core system upon initialization of the implementation.

    .. attribute:: _plugin_name

       The name of the plugin. Injected by the plugin core system upon initialization of the implementation.

    .. attribute:: _plugin_version

       The version of the plugin. Injected by the plugin core system upon initialization of the implementation.

    .. attribute:: _basefolder

       The base folder of the plugin. Injected by the plugin core system upon initialization of the implementation.

    .. attribute:: _logger

       The logger instance to use, with the logging name set to the :attr:`PluginManager.logging_prefix` of the
       :class:`PluginManager` concatenated with :attr:`_identifier`. Injected by the plugin core system upon
       initialization of the implementation.
    """

    def __init__(self):
        self._identifier = None
        self._plugin_name = None
        self._plugin_version = None
        self._basefolder = None
        self._logger = None

    def initialize(self):
        """
        Called by the plugin core after performing all injections. Override this to initialize your implementation.
        """
        pass

    def on_plugin_enabled(self):
        pass

    def on_plugin_disabled(self):
        pass


class RestartNeedingPlugin(Plugin):
    """
    Mixin for plugin types that need a restart after enabling/disabling them.
    """


class SortablePlugin(Plugin):
    """
    Mixin for plugin types that are sortable.
    """

    def get_sorting_key(self, context=None):
        """
        Returns the sorting key to use for the implementation in the specified ``context``.

        May return ``None`` if order is irrelevant.

        Implementations returning None will be ordered by plugin identifier
        after all implementations which did return a sorting key value that was
        not None sorted by that.

        Arguments:
            context (str): The sorting context for which to provide the
                sorting key value.

        Returns:
            int or None: An integer signifying the sorting key value of the plugin
                (sorting will be done ascending), or None if the implementation
                doesn't care about calling order.
        """
        return None


class PluginNeedsRestart(Exception):
    def __init__(self, name):
        Exception.__init__(self)
        self.name = name
        self.message = (
            "Plugin {name} cannot be enabled or disabled after system startup".format(
                **locals()
            )
        )


class PluginLifecycleException(Exception):
    def __init__(self, name, reason, message):
        Exception.__init__(self)
        self.name = name
        self.reason = reason

        self.message = message.format(**locals())

    def __str__(self):
        return self.message


class PluginCantInitialize(PluginLifecycleException):
    def __init__(self, name, reason):
        PluginLifecycleException.__init__(
            self, name, reason, "Plugin {name} cannot be initialized: {reason}"
        )


class PluginCantEnable(PluginLifecycleException):
    def __init__(self, name, reason):
        PluginLifecycleException.__init__(
            self, name, reason, "Plugin {name} cannot be enabled: {reason}"
        )


class PluginCantDisable(PluginLifecycleException):
    def __init__(self, name, reason):
        PluginLifecycleException.__init__(
            self, name, reason, "Plugin {name} cannot be disabled: {reason}"
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import click

click.disable_unicode_literals_warning = True

# ~~ "octoprint util" commands


@click.group()
def analysis_commands():
    pass


@analysis_commands.group(name="analysis")
def util():
    """Analysis tools."""
    pass


@util.command(name="gcode")
@click.option("--throttle", "throttle", type=float, default=None)
@click.option("--throttle-lines", "throttle_lines", type=int, default=None)
@click.option("--speed-x", "speedx", type=float, default=6000)
@click.option("--speed-y", "speedy", type=float, default=6000)
@click.option("--speed-z", "speedz", type=float, default=300)
@click.option("--offset", "offset", type=(float, float), multiple=True)
@click.option("--max-t", "maxt", type=int, default=10)
@click.option("--g90-extruder", "g90_extruder", is_flag=True)
@click.option("--progress", "progress", is_flag=True)
@click.argument("path", type=click.Path())
def gcode_command(
    path,
    speedx,
    speedy,
    speedz,
    offset,
    maxt,
    throttle,
    throttle_lines,
    g90_extruder,
    progress,
):
    """Runs a GCODE file analysis."""

    import time

    import yaml

    from octoprint.util import monotonic_time
    from octoprint.util.gcodeInterpreter import gcode

    throttle_callback = None
    if throttle:

        def throttle_callback(filePos, readBytes):
            if filePos % throttle_lines == 0:
                # only apply throttle every $throttle_lines lines
                time.sleep(throttle)

    offsets = offset
    if offsets is None:
        offsets = []
    elif isinstance(offset, tuple):
        offsets = list(offsets)
    offsets = [(0, 0)] + offsets
    if len(offsets) < maxt:
        offsets += [(0, 0)] * (maxt - len(offsets))

    start_time = monotonic_time()

    progress_callback = None
    if progress:

        def progress_callback(percentage):
            click.echo("PROGRESS:{}".format(percentage))

    interpreter = gcode(progress_callback=progress_callback)

    interpreter.load(
        path,
        speedx=speedx,
        speedy=speedy,
        offsets=offsets,
        throttle=throttle_callback,
        max_extruders=maxt,
        g90_extruder=g90_extruder,
    )

    click.echo("DONE:{}s".format(monotonic_time() - start_time))
    click.echo("RESULTS:")
    click.echo(
        yaml.safe_dump(
            interpreter.get_result(),
            default_flow_style=False,
            indent=4,
            allow_unicode=True,
        )
    )


if __name__ == "__main__":
    gcode_command()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from collections import deque

from tornado import ioloop


class MovingAverage(object):
    """Moving average class implementation"""
    def __init__(self, period=10):
        """Constructor.

        `period`
            Moving window size. Average will be calculated
            from the data in the window.
        """
        self.period = period
        self.stream = deque()
        self.sum = 0
        self.accumulator = 0
        self.last_average = 0

    def add(self, n):
        """Add value to the current accumulator

        `n`
            Value to add
        """
        self.accumulator += n

    def flush(self):
        """Add accumulator to the moving average queue
        and reset it. For example, called by the StatsCollector
        once per second to calculate per-second average.
        """
        n = self.accumulator
        self.accumulator = 0

        stream = self.stream
        stream.append(n)
        self.sum += n

        streamlen = len(stream)

        if streamlen > self.period:
            self.sum -= stream.popleft()
            streamlen -= 1

        if streamlen == 0:
            self.last_average = 0
        else:
            self.last_average = self.sum / float(streamlen)


class StatsCollector(object):
    def __init__(self):
        # Sessions
        self.sess_active = 0

        # Avoid circular reference
        self.sess_transports = {}

        # Connections
        self.conn_active = 0
        self.conn_ps = MovingAverage()

        # Packets
        self.pack_sent_ps = MovingAverage()
        self.pack_recv_ps = MovingAverage()

        self._callback = ioloop.PeriodicCallback(self._update,
                                                 1000)
        self._callback.start()

    def _update(self):
        self.conn_ps.flush()

        self.pack_sent_ps.flush()
        self.pack_recv_ps.flush()

    def dump(self):
        """Return dictionary with current statistical information"""
        data = {"sessions_active": self.sess_active,

            # Connections
        "connections_active": self.conn_active,
        "connections_ps": self.conn_ps.last_average,

            # Packets
        "packets_sent_ps": self.pack_sent_ps.last_average,
        "packets_recv_ps": self.pack_recv_ps.last_average}

        for k, v in self.sess_transports.items():
            data['transp_' + k] = v

        return data

    # Various event callbacks
    def on_sess_opened(self, transport):
        self.sess_active += 1

        if transport not in self.sess_transports:
            self.sess_transports[transport] = 0

        self.sess_transports[transport] += 1

    def on_sess_closed(self, transport):
        self.sess_active -= 1
        self.sess_transports[transport] -= 1

    def on_conn_opened(self):
        self.conn_active += 1
        self.conn_ps.add(1)

    def on_conn_closed(self):
        self.conn_active -= 1

    def on_pack_sent(self, num):
        self.pack_sent_ps.add(num)

    def on_pack_recv(self, num):
        self.pack_recv_ps.add(num)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

import io
import os

import flask
import sarge
from flask_babel import gettext

import octoprint.events
import octoprint.plugin
from octoprint.access.groups import USER_GROUP
from octoprint.access.permissions import Permissions
from octoprint.util import RepeatedTimer
from octoprint.util.platform import CLOSE_FDS

_PROC_DT_MODEL_PATH = "/proc/device-tree/model"
_OCTOPI_VERSION_PATH = "/etc/octopi_version"
_VCGENCMD_THROTTLE = "/usr/bin/vcgencmd get_throttled"

_CHECK_INTERVAL_OK = 300
_CHECK_INTERVAL_THROTTLED = 30

__LOCAL_DEBUG = os.path.exists(
    os.path.realpath(os.path.join(os.path.dirname(__file__), ".local_debug"))
)

if __LOCAL_DEBUG:
    ### mocks & settings for local debugging
    import sys

    base = os.path.realpath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            "..",
            "..",
            "tests",
            "plugins",
            "pi_support",
            "fakes",
        )
    )
    _PROC_DT_MODEL_PATH = os.path.join(base, "fake_model.txt")
    _OCTOPI_VERSION_PATH = os.path.join(base, "fake_octopi.txt")
    _VCGENCMD_THROTTLE = "{} {}".format(
        sys.executable, os.path.join(base, "fake_vcgencmd.py")
    )
    import itertools

    _VCGENCMD_OUTPUT = itertools.chain(
        iter(("0x0", "0x0", "0x50005", "0x50000", "0x70007")), itertools.repeat("0x70005")
    )

    _CHECK_INTERVAL_OK = 10
    _CHECK_INTERVAL_THROTTLED = 5


# see https://www.raspberrypi.org/forums/viewtopic.php?f=63&t=147781&start=50#p972790
_FLAG_UNDERVOLTAGE = 1 << 0
_FLAG_FREQ_CAPPED = 1 << 1
_FLAG_THROTTLED = 1 << 2
_FLAG_PAST_UNDERVOLTAGE = 1 << 16
_FLAG_PAST_FREQ_CAPPED = 1 << 17
_FLAG_PAST_THROTTLED = 1 << 18


class ThrottleState(object):
    @classmethod
    def from_value(cls, value=0):
        if value == 0:
            return ThrottleState(raw_value=value)

        kwargs = {
            "undervoltage": _FLAG_UNDERVOLTAGE & value == _FLAG_UNDERVOLTAGE,
            "freq_capped": _FLAG_FREQ_CAPPED & value == _FLAG_FREQ_CAPPED,
            "throttled": _FLAG_THROTTLED & value == _FLAG_THROTTLED,
            "past_undervoltage": _FLAG_PAST_UNDERVOLTAGE & value
            == _FLAG_PAST_UNDERVOLTAGE,
            "past_freq_capped": _FLAG_PAST_FREQ_CAPPED & value == _FLAG_PAST_FREQ_CAPPED,
            "past_throttled": _FLAG_PAST_THROTTLED & value == _FLAG_PAST_THROTTLED,
            "raw_value": value,
        }
        return ThrottleState(**kwargs)

    def __init__(self, **kwargs):
        self._raw_value = kwargs.get("raw_value", -1)
        self._undervoltage = False
        self._freq_capped = False
        self._throttled = False
        self._past_undervoltage = False
        self._past_freq_capped = False
        self._past_throttled = False

        for key, value in kwargs.items():
            local_key = "_{}".format(key)
            if hasattr(self, local_key) and isinstance(value, bool):
                setattr(self, local_key, value)

    @property
    def undervoltage(self):
        return self.current_undervoltage or self.past_undervoltage

    @property
    def current_undervoltage(self):
        return self._undervoltage

    @property
    def past_undervoltage(self):
        return self._past_undervoltage

    @property
    def overheat(self):
        return self.current_overheat or self.past_overheat

    @property
    def current_overheat(self):
        return self._freq_capped

    @property
    def past_overheat(self):
        return self._past_freq_capped

    @property
    def issue(self):
        return self.current_issue or self.past_issue

    @property
    def current_issue(self):
        return self._undervoltage or self._freq_capped or self._throttled

    @property
    def past_issue(self):
        return self._past_undervoltage or self._past_freq_capped or self._past_throttled

    @property
    def raw_value(self):
        return self._raw_value

    @property
    def raw_value_hex(self):
        return "0x{:X}".format(self._raw_value)

    def __eq__(self, other):
        if not isinstance(other, ThrottleState):
            return False

        return (
            self._undervoltage == other._undervoltage
            and self._freq_capped == other._freq_capped
            and self._throttled == other._throttled
            and self._past_undervoltage == other._past_undervoltage
            and self._past_freq_capped == other._past_freq_capped
            and self._past_throttled == other._past_throttled
            and self._raw_value == other._raw_value
        )

    def as_dict(self):
        return {
            "raw_value": self._raw_value,
            "current_undervoltage": self.current_undervoltage,
            "past_undervoltage": self.past_undervoltage,
            "current_overheat": self.current_overheat,
            "past_overheat": self.past_overheat,
            "current_issue": self.current_issue,
            "past_issue": self.past_issue,
        }


_proc_dt_model = None


def get_proc_dt_model():
    global _proc_dt_model

    if _proc_dt_model is None:
        with io.open(_PROC_DT_MODEL_PATH, "rt", encoding="utf-8") as f:
            _proc_dt_model = f.readline().strip(" \t\r\n\0")

    return _proc_dt_model


def get_vcgencmd_throttled_state(command):
    if __LOCAL_DEBUG:
        output = "throttled={}".format(next(_VCGENCMD_OUTPUT))  # mock for local debugging
    else:
        output = sarge.get_stdout(command, close_fds=CLOSE_FDS)

    if "throttled=0x" not in output:
        raise ValueError('cannot parse "{}" output: {}'.format(command, output))

    value = output[len("throttled=") :].strip(" \t\r\n\0")
    value = int(value, 0)
    return ThrottleState.from_value(value)


def is_octopi():
    return os.path.exists(_OCTOPI_VERSION_PATH)


_octopi_version = None


def get_octopi_version():
    global _octopi_version

    if _octopi_version is None:
        with io.open(_OCTOPI_VERSION_PATH, "rt", encoding="utf-8") as f:
            _octopi_version = f.readline().strip(" \t\r\n\0")

    return _octopi_version


class PiSupportPlugin(
    octoprint.plugin.EnvironmentDetectionPlugin,
    octoprint.plugin.SimpleApiPlugin,
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.StartupPlugin,
    octoprint.plugin.SettingsPlugin,
):

    # noinspection PyMissingConstructor
    def __init__(self):
        self._throttle_state = ThrottleState()
        self._throttle_check = None
        self._throttle_undervoltage = False
        self._throttle_overheat = False
        self._throttle_functional = True

    # Additional permissions hook

    def get_additional_permissions(self):
        return [
            {
                "key": "STATUS",
                "name": "Status",
                "description": gettext(
                    "Allows to check for the Pi's throttling status and environment info"
                ),
                "roles": ["check"],
                "default_groups": [USER_GROUP],
            }
        ]

    # ~~ EnvironmentDetectionPlugin

    def get_additional_environment(self):
        result = {"model": get_proc_dt_model()}

        self._check_throttled_state()
        if self._throttle_functional:
            result["throttle_state"] = self._throttle_state.raw_value_hex

        if is_octopi():
            result["octopi_version"] = get_octopi_version()

        return result

    # ~~ SimpleApiPlugin

    def on_api_get(self, request):
        if not Permissions.PLUGIN_PI_SUPPORT_STATUS.can():
            return flask.abort(403)

        environment = self.get_additional_environment()

        result = {
            "throttle_state": self._throttle_state.as_dict(),
            "model_unrecommended": "zero" in environment.get("model").lower(),
        }
        result.update(environment)
        return flask.jsonify(**result)

    # ~~ AssetPlugin

    def get_assets(self):
        return {
            "js": ["js/pi_support.js"],
            "clientjs": ["clientjs/pi_support.js"],
            "css": ["css/pi_support.css"],
        }

    # ~~ TemplatePlugin

    def get_template_configs(self):
        configs = [
            {
                "type": "settings",
                "name": gettext("Pi Support"),
                "template": "pi_support_settings.jinja2",
                "custom_bindings": False,
            }
        ]

        if is_octopi():
            configs.append(
                {
                    "type": "about",
                    "name": "About OctoPi",
                    "template": "pi_support_about_octopi.jinja2",
                }
            )

        return configs

    def get_template_vars(self):
        return self.get_additional_environment()

    # ~~ StartupPlugin

    def on_startup(self, *args, **kwargs):
        if self._settings.get_boolean(["vcgencmd_throttle_check_enabled"]):
            self._check_throttled_state()
            self._throttle_check = RepeatedTimer(
                self._check_throttled_state_interval,
                self._check_throttled_state,
                condition=self._check_throttled_state_condition,
            )
            self._throttle_check.start()

    # ~~ SettingsPlugin

    def get_settings_defaults(self):
        return {
            "vcgencmd_throttle_check_enabled": True,
            "vcgencmd_throttle_check_command": _VCGENCMD_THROTTLE,
            "ignore_unrecommended_model": False,
        }

    def get_settings_restricted_paths(self):
        return {
            "admin": [
                ["vcgencmd_throttle_check_enabled"],
                ["vcgencmd_throttle_check_command"],
            ]
        }

    # ~~ Helpers

    def _check_throttled_state_interval(self):
        if self._throttle_state.current_issue:
            return _CHECK_INTERVAL_THROTTLED
        else:
            return _CHECK_INTERVAL_OK

    def _check_throttled_state_condition(self):
        return self._throttle_functional

    def get_throttle_state(self, run_now=False):
        """Exposed as public helper."""
        if run_now:
            self._check_throttled_state()

        if not self._throttle_functional:
            return False

        return self._throttle_state.as_dict()

    def _check_throttled_state(self):
        command = self._settings.get(["vcgencmd_throttle_check_command"])

        self._logger.debug('Retrieving throttle state via "{}"'.format(command))
        try:
            state = get_vcgencmd_throttled_state(command)
        except ValueError:
            self._logger.warning(
                'Fetching the current throttle state via "{}" doesn\'t work'.format(
                    command
                )
            )
            self._throttle_functional = False
            return

        if self._throttle_state == state:
            # no change
            return

        self._throttle_state = state

        if (not self._throttle_undervoltage and self._throttle_state.undervoltage) or (
            not self._throttle_overheat and self._throttle_state.overheat
        ):
            message = (
                "This Raspberry Pi is reporting problems that might lead to bad performance or errors caused "
                "by overheating or insufficient power."
            )

            if self._throttle_state.undervoltage:
                self._throttle_undervoltage = True
                message += (
                    "\n!!! UNDERVOLTAGE REPORTED !!! Make sure that the power supply and power cable are "
                    "capable of supplying enough voltage and current to your Pi."
                )

            if self._throttle_state.overheat:
                self._throttle_overheat = True
                message += (
                    "\n!!! FREQUENCY CAPPING DUE TO OVERHEATING REPORTED !!! Improve cooling on the Pi's "
                    "CPU and GPU."
                )

            self._logger.warning(message)

        self._plugin_manager.send_plugin_message(
            self._identifier,
            {"type": "throttle_state", "state": self._throttle_state.as_dict()},
        )

        # noinspection PyUnresolvedReferences
        self._event_bus.fire(
            octoprint.events.Events.PLUGIN_PI_SUPPORT_THROTTLE_STATE,
            self._throttle_state.as_dict(),
        )


def register_custom_events(*args, **kwargs):
    return [
        "throttle_state",
    ]


__plugin_name__ = "Pi Support Plugin"
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Provides additional information about your Pi in the UI."
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin OctoPrint will no longer be able to "
    "provide additional information about your Pi, "
    "which will make it more tricky to help you if you need support."
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=2.7,<4"


def __plugin_check__():
    try:
        proc_dt_model = get_proc_dt_model()
        if proc_dt_model is None:
            return False
    except Exception:
        return False

    return "raspberry pi" in proc_dt_model.lower()


def __plugin_load__():
    plugin = PiSupportPlugin()
    global __plugin_implementation__
    __plugin_implementation__ = plugin

    global __plugin_hooks__
    __plugin_hooks__ = {
        "octoprint.events.register_custom_events": register_custom_events,
        "octoprint.access.permissions": __plugin_implementation__.get_additional_permissions,
    }

    global __plugin_helpers__
    __plugin_helpers__ = {"get_throttled": plugin.get_throttle_state}



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

import hashlib
import logging
import os
import threading

import psutil
from flask import jsonify, make_response, request, url_for

import octoprint.filemanager
import octoprint.filemanager.storage
import octoprint.filemanager.util
import octoprint.slicing
from octoprint.access.permissions import Permissions
from octoprint.events import Events
from octoprint.filemanager.destinations import FileDestinations
from octoprint.server import (
    NO_CONTENT,
    current_user,
    eventManager,
    fileManager,
    printer,
    slicingManager,
)
from octoprint.server.api import api
from octoprint.server.util.flask import (
    get_json_command_from_request,
    no_firstrun_access,
    with_revalidation_checking,
)
from octoprint.settings import settings, valid_boolean_trues
from octoprint.util import sv, time_this

# ~~ GCODE file handling

_file_cache = {}
_file_cache_mutex = threading.RLock()

_DATA_FORMAT_VERSION = "v2"


def _clear_file_cache():
    with _file_cache_mutex:
        _file_cache.clear()


def _create_lastmodified(path, recursive):
    path = path[len("/api/files") :]
    if path.startswith("/"):
        path = path[1:]

    if path == "":
        # all storages involved
        lms = [0]
        for storage in fileManager.registered_storages:
            try:
                lms.append(fileManager.last_modified(storage, recursive=recursive))
            except Exception:
                logging.getLogger(__name__).exception(
                    "There was an error retrieving the last modified data from storage {}".format(
                        storage
                    )
                )
                lms.append(None)

        if any(filter(lambda x: x is None, lms)):
            # we return None if ANY of the involved storages returned None
            return None

        # if we reach this point, we return the maximum of all dates
        return max(lms)

    else:
        if "/" in path:
            storage, path_in_storage = path.split("/", 1)
        else:
            storage = path
            path_in_storage = None

        try:
            return fileManager.last_modified(
                storage, path=path_in_storage, recursive=recursive
            )
        except Exception:
            logging.getLogger(__name__).exception(
                "There was an error retrieving the last modified data from storage {} and path {}".format(
                    storage, path_in_storage
                )
            )
            return None


def _create_etag(path, filter, recursive, lm=None):
    if lm is None:
        lm = _create_lastmodified(path, recursive)

    if lm is None:
        return None

    hash = hashlib.sha1()

    def hash_update(value):
        value = value.encode("utf-8")
        hash.update(value)

    hash_update(str(lm))
    hash_update(str(filter))
    hash_update(str(recursive))

    path = path[len("/api/files") :]
    if path.startswith("/"):
        path = path[1:]

    if "/" in path:
        storage, _ = path.split("/", 1)
    else:
        storage = path

    if path == "" or storage == FileDestinations.SDCARD:
        # include sd data in etag
        hash_update(repr(sorted(printer.get_sd_files(), key=lambda x: sv(x["name"]))))

    hash_update(_DATA_FORMAT_VERSION)  # increment version if we change the API format

    return hash.hexdigest()


@api.route("/files", methods=["GET"])
@Permissions.FILES_LIST.require(403)
@with_revalidation_checking(
    etag_factory=lambda lm=None: _create_etag(
        request.path,
        request.values.get("filter", False),
        request.values.get("recursive", False),
        lm=lm,
    ),
    lastmodified_factory=lambda: _create_lastmodified(
        request.path, request.values.get("recursive", False)
    ),
    unless=lambda: request.values.get("force", False)
    or request.values.get("_refresh", False),
)
def readGcodeFiles():
    filter = request.values.get("filter", False)
    recursive = request.values.get("recursive", "false") in valid_boolean_trues
    force = request.values.get("force", "false") in valid_boolean_trues

    files = _getFileList(
        FileDestinations.LOCAL,
        filter=filter,
        recursive=recursive,
        allow_from_cache=not force,
    )
    files.extend(_getFileList(FileDestinations.SDCARD))

    usage = psutil.disk_usage(settings().getBaseFolder("uploads", check_writable=False))
    return jsonify(files=files, free=usage.free, total=usage.total)


@api.route("/files/<string:origin>", methods=["GET"])
@Permissions.FILES_LIST.require(403)
@with_revalidation_checking(
    etag_factory=lambda lm=None: _create_etag(
        request.path,
        request.values.get("filter", False),
        request.values.get("recursive", False),
        lm=lm,
    ),
    lastmodified_factory=lambda: _create_lastmodified(
        request.path, request.values.get("recursive", False)
    ),
    unless=lambda: request.values.get("force", False)
    or request.values.get("_refresh", False),
)
def readGcodeFilesForOrigin(origin):
    if origin not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
        return make_response("Unknown origin: %s" % origin, 404)

    filter = request.values.get("filter", False)
    recursive = request.values.get("recursive", "false") in valid_boolean_trues
    force = request.values.get("force", "false") in valid_boolean_trues

    files = _getFileList(
        origin, filter=filter, recursive=recursive, allow_from_cache=not force
    )

    if origin == FileDestinations.LOCAL:
        usage = psutil.disk_usage(
            settings().getBaseFolder("uploads", check_writable=False)
        )
        return jsonify(files=files, free=usage.free, total=usage.total)
    else:
        return jsonify(files=files)


@api.route("/files/<string:target>/<path:filename>", methods=["GET"])
@Permissions.FILES_LIST.require(403)
@with_revalidation_checking(
    etag_factory=lambda lm=None: _create_etag(
        request.path,
        request.values.get("filter", False),
        request.values.get("recursive", False),
        lm=lm,
    ),
    lastmodified_factory=lambda: _create_lastmodified(
        request.path, request.values.get("recursive", False)
    ),
    unless=lambda: request.values.get("force", False)
    or request.values.get("_refresh", False),
)
def readGcodeFile(target, filename):
    if target not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
        return make_response("Unknown target: %s" % target, 404)

    recursive = False
    if "recursive" in request.values:
        recursive = request.values["recursive"] in valid_boolean_trues

    file = _getFileDetails(target, filename, recursive=recursive)
    if not file:
        return make_response("File not found on '%s': %s" % (target, filename), 404)

    return jsonify(file)


def _getFileDetails(origin, path, recursive=True):
    parent, path = os.path.split(path)
    files = _getFileList(origin, path=parent, recursive=recursive, level=1)

    for f in files:
        if f["name"] == path:
            return f
    else:
        return None


@time_this(
    logtarget=__name__ + ".timings",
    message="{func}({func_args},{func_kwargs}) took {timing:.2f}ms",
    incl_func_args=True,
)
def _getFileList(
    origin, path=None, filter=None, recursive=False, level=0, allow_from_cache=True
):
    if origin == FileDestinations.SDCARD:
        sdFileList = printer.get_sd_files()

        files = []
        if sdFileList is not None:
            for f in sdFileList:
                type_path = octoprint.filemanager.get_file_type(f["name"])
                if not type_path:
                    # only supported extensions
                    continue
                else:
                    file_type = type_path[0]

                file = {
                    "type": file_type,
                    "typePath": type_path,
                    "name": f["name"],
                    "display": f["display"] if f["display"] else f["name"],
                    "path": f["name"],
                    "origin": FileDestinations.SDCARD,
                    "refs": {
                        "resource": url_for(
                            ".readGcodeFile",
                            target=FileDestinations.SDCARD,
                            filename=f["name"],
                            _external=True,
                        )
                    },
                }
                if f["size"] is not None:
                    file.update({"size": f["size"]})
                files.append(file)
    else:
        filter_func = None
        if filter:
            filter_func = lambda entry, entry_data: octoprint.filemanager.valid_file_type(
                entry, type=filter
            )

        with _file_cache_mutex:
            cache_key = "{}:{}:{}:{}".format(origin, path, recursive, filter)
            files, lastmodified = _file_cache.get(cache_key, ([], None))
            # recursive needs to be True for lastmodified queries so we get lastmodified of whole subtree - #3422
            if (
                not allow_from_cache
                or lastmodified is None
                or lastmodified
                < fileManager.last_modified(origin, path=path, recursive=True)
            ):
                files = list(
                    fileManager.list_files(
                        origin,
                        path=path,
                        filter=filter_func,
                        recursive=recursive,
                        level=level,
                        force_refresh=not allow_from_cache,
                    )[origin].values()
                )
                lastmodified = fileManager.last_modified(
                    origin, path=path, recursive=True
                )
                _file_cache[cache_key] = (files, lastmodified)

        def analyse_recursively(files, path=None):
            if path is None:
                path = ""

            result = []
            for file_or_folder in files:
                # make a shallow copy in order to not accidentally modify the cached data
                file_or_folder = dict(file_or_folder)

                file_or_folder["origin"] = FileDestinations.LOCAL

                if file_or_folder["type"] == "folder":
                    if "children" in file_or_folder:
                        file_or_folder["children"] = analyse_recursively(
                            file_or_folder["children"].values(),
                            path + file_or_folder["name"] + "/",
                        )

                    file_or_folder["refs"] = {
                        "resource": url_for(
                            ".readGcodeFile",
                            target=FileDestinations.LOCAL,
                            filename=path + file_or_folder["name"],
                            _external=True,
                        )
                    }
                else:
                    if (
                        "analysis" in file_or_folder
                        and octoprint.filemanager.valid_file_type(
                            file_or_folder["name"], type="gcode"
                        )
                    ):
                        file_or_folder["gcodeAnalysis"] = file_or_folder["analysis"]
                        del file_or_folder["analysis"]

                    if (
                        "history" in file_or_folder
                        and octoprint.filemanager.valid_file_type(
                            file_or_folder["name"], type="gcode"
                        )
                    ):
                        # convert print log
                        history = file_or_folder["history"]
                        del file_or_folder["history"]
                        success = 0
                        failure = 0
                        last = None
                        for entry in history:
                            success += 1 if "success" in entry and entry["success"] else 0
                            failure += (
                                1 if "success" in entry and not entry["success"] else 0
                            )
                            if not last or (
                                "timestamp" in entry
                                and "timestamp" in last
                                and entry["timestamp"] > last["timestamp"]
                            ):
                                last = entry
                        if last:
                            prints = {
                                "success": success,
                                "failure": failure,
                                "last": {
                                    "success": last["success"],
                                    "date": last["timestamp"],
                                },
                            }
                            if "printTime" in last:
                                prints["last"]["printTime"] = last["printTime"]
                            file_or_folder["prints"] = prints

                    file_or_folder["refs"] = {
                        "resource": url_for(
                            ".readGcodeFile",
                            target=FileDestinations.LOCAL,
                            filename=file_or_folder["path"],
                            _external=True,
                        ),
                        "download": url_for("index", _external=True)
                        + "downloads/files/"
                        + FileDestinations.LOCAL
                        + "/"
                        + file_or_folder["path"],
                    }

                result.append(file_or_folder)

            return result

        files = analyse_recursively(files)

    return files


def _verifyFileExists(origin, filename):
    if origin == FileDestinations.SDCARD:
        return filename in (x["name"] for x in printer.get_sd_files())
    else:
        return fileManager.file_exists(origin, filename)


def _verifyFolderExists(origin, foldername):
    if origin == FileDestinations.SDCARD:
        return False
    else:
        return fileManager.folder_exists(origin, foldername)


def _isBusy(target, path):
    currentOrigin, currentPath = _getCurrentFile()
    if (
        currentPath is not None
        and currentOrigin == target
        and fileManager.file_in_path(FileDestinations.LOCAL, path, currentPath)
        and (printer.is_printing() or printer.is_paused())
    ):
        return True

    return any(
        target == x[0] and fileManager.file_in_path(FileDestinations.LOCAL, path, x[1])
        for x in fileManager.get_busy_files()
    )


@api.route("/files/<string:target>", methods=["POST"])
@no_firstrun_access
@Permissions.FILES_UPLOAD.require(403)
def uploadGcodeFile(target):
    input_name = "file"
    input_upload_name = (
        input_name + "." + settings().get(["server", "uploads", "nameSuffix"])
    )
    input_upload_path = (
        input_name + "." + settings().get(["server", "uploads", "pathSuffix"])
    )
    if input_upload_name in request.values and input_upload_path in request.values:
        if target not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
            return make_response("Unknown target: %s" % target, 404)

        upload = octoprint.filemanager.util.DiskFileWrapper(
            request.values[input_upload_name], request.values[input_upload_path]
        )

        # Store any additional user data the caller may have passed.
        userdata = None
        if "userdata" in request.values:
            import json

            try:
                userdata = json.loads(request.values["userdata"])
            except Exception:
                return make_response("userdata contains invalid JSON", 400)

        if target == FileDestinations.SDCARD and not settings().getBoolean(
            ["feature", "sdSupport"]
        ):
            return make_response("SD card support is disabled", 404)

        sd = target == FileDestinations.SDCARD
        selectAfterUpload = (
            "select" in request.values
            and request.values["select"] in valid_boolean_trues
            and Permissions.FILES_SELECT.can()
        )
        printAfterSelect = (
            "print" in request.values
            and request.values["print"] in valid_boolean_trues
            and Permissions.PRINT.can()
        )

        if sd:
            # validate that all preconditions for SD upload are met before attempting it
            if not (
                printer.is_operational()
                and not (printer.is_printing() or printer.is_paused())
            ):
                return make_response(
                    "Can not upload to SD card, printer is either not operational or already busy",
                    409,
                )
            if not printer.is_sd_ready():
                return make_response(
                    "Can not upload to SD card, not yet initialized", 409
                )

        # determine future filename of file to be uploaded, abort if it can't be uploaded
        try:
            # FileDestinations.LOCAL = should normally be target, but can't because SDCard handling isn't implemented yet
            canonPath, canonFilename = fileManager.canonicalize(
                FileDestinations.LOCAL, upload.filename
            )
            futurePath = fileManager.sanitize_path(FileDestinations.LOCAL, canonPath)
            futureFilename = fileManager.sanitize_name(
                FileDestinations.LOCAL, canonFilename
            )
        except Exception:
            canonFilename = None
            futurePath = None
            futureFilename = None

        if futureFilename is None:
            return make_response(
                "Can not upload file %s, wrong format?" % upload.filename, 415
            )

        if "path" in request.values and request.values["path"]:
            # we currently only support uploads to sdcard via local, so first target is local instead of "target"
            futurePath = fileManager.sanitize_path(
                FileDestinations.LOCAL, request.values["path"]
            )

        # prohibit overwriting currently selected file while it's being printed
        futureFullPath = fileManager.join_path(
            FileDestinations.LOCAL, futurePath, futureFilename
        )
        futureFullPathInStorage = fileManager.path_in_storage(
            FileDestinations.LOCAL, futureFullPath
        )

        if not printer.can_modify_file(futureFullPathInStorage, sd):
            return make_response(
                "Trying to overwrite file that is currently being printed: %s"
                % futureFullPath,
                409,
            )

        reselect = printer.is_current_file(futureFullPathInStorage, sd)

        user = current_user.get_name()

        def fileProcessingFinished(filename, absFilename, destination):
            """
            Callback for when the file processing (upload, optional slicing, addition to analysis queue) has
            finished.

            Depending on the file's destination triggers either streaming to SD card or directly calls selectAndOrPrint.
            """

            if (
                destination == FileDestinations.SDCARD
                and octoprint.filemanager.valid_file_type(filename, "machinecode")
            ):
                return filename, printer.add_sd_file(
                    filename,
                    absFilename,
                    on_success=selectAndOrPrint,
                    tags={"source:api", "api:files.sd"},
                )
            else:
                selectAndOrPrint(filename, absFilename, destination)
                return filename

        def selectAndOrPrint(filename, absFilename, destination):
            """
            Callback for when the file is ready to be selected and optionally printed. For SD file uploads this is only
            the case after they have finished streaming to the printer, which is why this callback is also used
            for the corresponding call to addSdFile.

            Selects the just uploaded file if either selectAfterUpload or printAfterSelect are True, or if the
            exact file is already selected, such reloading it.
            """
            if octoprint.filemanager.valid_file_type(added_file, "gcode") and (
                selectAfterUpload or printAfterSelect or reselect
            ):
                printer.select_file(
                    absFilename,
                    destination == FileDestinations.SDCARD,
                    printAfterSelect,
                    user,
                )

        try:
            added_file = fileManager.add_file(
                FileDestinations.LOCAL,
                futureFullPathInStorage,
                upload,
                allow_overwrite=True,
                display=canonFilename,
            )
        except octoprint.filemanager.storage.StorageError as e:
            if e.code == octoprint.filemanager.storage.StorageError.INVALID_FILE:
                return make_response(
                    'Could not upload the file "{}", invalid type'.format(
                        upload.filename
                    ),
                    400,
                )
            else:
                return make_response(
                    'Could not upload the file "{}"'.format(upload.filename), 500
                )

        if octoprint.filemanager.valid_file_type(added_file, "stl"):
            filename = added_file
            done = True
        else:
            filename = fileProcessingFinished(
                added_file,
                fileManager.path_on_disk(FileDestinations.LOCAL, added_file),
                target,
            )
            done = not sd

        if userdata is not None:
            # upload included userdata, add this now to the metadata
            fileManager.set_additional_metadata(
                FileDestinations.LOCAL, added_file, "userdata", userdata
            )

        sdFilename = None
        if isinstance(filename, tuple):
            filename, sdFilename = filename

        payload = {
            "name": futureFilename,
            "path": filename,
            "target": target,
            "select": selectAfterUpload,
            "print": printAfterSelect,
        }
        if userdata is not None:
            payload["userdata"] = userdata
        eventManager.fire(Events.UPLOAD, payload)

        files = {}
        location = url_for(
            ".readGcodeFile",
            target=FileDestinations.LOCAL,
            filename=filename,
            _external=True,
        )
        files.update(
            {
                FileDestinations.LOCAL: {
                    "name": futureFilename,
                    "path": filename,
                    "origin": FileDestinations.LOCAL,
                    "refs": {
                        "resource": location,
                        "download": url_for("index", _external=True)
                        + "downloads/files/"
                        + FileDestinations.LOCAL
                        + "/"
                        + filename,
                    },
                }
            }
        )

        if sd and sdFilename:
            location = url_for(
                ".readGcodeFile",
                target=FileDestinations.SDCARD,
                filename=sdFilename,
                _external=True,
            )
            files.update(
                {
                    FileDestinations.SDCARD: {
                        "name": sdFilename,
                        "path": sdFilename,
                        "origin": FileDestinations.SDCARD,
                        "refs": {"resource": location},
                    }
                }
            )

        r = make_response(jsonify(files=files, done=done), 201)
        r.headers["Location"] = location
        return r

    elif "foldername" in request.values:
        foldername = request.values["foldername"]

        if target not in [FileDestinations.LOCAL]:
            return make_response("Unknown target: %s" % target, 400)

        canonPath, canonName = fileManager.canonicalize(target, foldername)
        futurePath = fileManager.sanitize_path(target, canonPath)
        futureName = fileManager.sanitize_name(target, canonName)
        if not futureName or not futurePath:
            return make_response("Can't create a folder with an empty name", 400)

        if "path" in request.values and request.values["path"]:
            futurePath = fileManager.sanitize_path(
                FileDestinations.LOCAL, request.values["path"]
            )

        futureFullPath = fileManager.join_path(target, futurePath, futureName)
        if octoprint.filemanager.valid_file_type(futureName):
            return make_response(
                "Can't create a folder named %s, please try another name" % futureName,
                409,
            )

        try:
            added_folder = fileManager.add_folder(
                target, futureFullPath, display=canonName
            )
        except octoprint.filemanager.storage.StorageError as e:
            if e.code == octoprint.filemanager.storage.StorageError.INVALID_DIRECTORY:
                return make_response(
                    "Could not create folder {}, invalid directory".format(futureName)
                )
            else:
                return make_response("Could not create folder {}".format(futureName))

        location = url_for(
            ".readGcodeFile",
            target=FileDestinations.LOCAL,
            filename=added_folder,
            _external=True,
        )
        folder = {
            "name": futureName,
            "path": added_folder,
            "origin": target,
            "refs": {"resource": location},
        }

        r = make_response(jsonify(folder=folder, done=True), 201)
        r.headers["Location"] = location
        return r

    else:
        return make_response("No file to upload and no folder to create", 400)


@api.route("/files/<string:target>/<path:filename>", methods=["POST"])
@no_firstrun_access
def gcodeFileCommand(filename, target):
    if target not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
        return make_response("Unknown target: %s" % target, 404)

    # valid file commands, dict mapping command name to mandatory parameters
    valid_commands = {
        "select": [],
        "slice": [],
        "analyse": [],
        "copy": ["destination"],
        "move": ["destination"],
    }

    command, data, response = get_json_command_from_request(request, valid_commands)
    if response is not None:
        return response

    user = current_user.get_name()

    if command == "select":
        with Permissions.FILES_SELECT.require(403):
            if not _verifyFileExists(target, filename):
                return make_response(
                    "File not found on '%s': %s" % (target, filename), 404
                )

            # selects/loads a file
            if not octoprint.filemanager.valid_file_type(filename, type="machinecode"):
                return make_response(
                    "Cannot select {filename} for printing, not a machinecode file".format(
                        **locals()
                    ),
                    415,
                )

            if not printer.is_ready():
                return make_response(
                    "Printer is already printing, cannot select a new file", 409
                )

            printAfterLoading = False
            if "print" in data and data["print"] in valid_boolean_trues:
                with Permissions.PRINT.require(403):
                    if not printer.is_operational():
                        return make_response(
                            "Printer is not operational, cannot directly start printing",
                            409,
                        )
                    printAfterLoading = True

            sd = False
            if target == FileDestinations.SDCARD:
                filenameToSelect = filename
                sd = True
            else:
                filenameToSelect = fileManager.path_on_disk(target, filename)
            printer.select_file(filenameToSelect, sd, printAfterLoading, user)

    elif command == "slice":
        with Permissions.SLICE.require(403):
            if not _verifyFileExists(target, filename):
                return make_response(
                    "File not found on '%s': %s" % (target, filename), 404
                )

            try:
                if "slicer" in data:
                    slicer = data["slicer"]
                    del data["slicer"]
                    slicer_instance = slicingManager.get_slicer(slicer)

                elif "cura" in slicingManager.registered_slicers:
                    slicer = "cura"
                    slicer_instance = slicingManager.get_slicer("cura")

                else:
                    return make_response(
                        "Cannot slice {filename}, no slicer available".format(**locals()),
                        415,
                    )
            except octoprint.slicing.UnknownSlicer as e:
                return make_response(
                    "Slicer {slicer} is not available".format(slicer=e.slicer), 400
                )

            if not any(
                [
                    octoprint.filemanager.valid_file_type(filename, type=source_file_type)
                    for source_file_type in slicer_instance.get_slicer_properties().get(
                        "source_file_types", ["model"]
                    )
                ]
            ):
                return make_response(
                    "Cannot slice {filename}, not a model file".format(**locals()), 415
                )

            cores = psutil.cpu_count()
            if (
                slicer_instance.get_slicer_properties().get("same_device", True)
                and (printer.is_printing() or printer.is_paused())
                and (cores is None or cores < 2)
            ):
                # slicer runs on same device as OctoPrint, slicing while printing is hence disabled
                return make_response(
                    "Cannot slice on {slicer} while printing on single core systems or systems of unknown core count due to performance reasons".format(
                        **locals()
                    ),
                    409,
                )

            if "destination" in data and data["destination"]:
                destination = data["destination"]
                del data["destination"]
            elif "gcode" in data and data["gcode"]:
                destination = data["gcode"]
                del data["gcode"]
            else:
                import os

                name, _ = os.path.splitext(filename)
                destination = (
                    name
                    + "."
                    + slicer_instance.get_slicer_properties().get(
                        "destination_extensions", ["gco", "gcode", "g"]
                    )[0]
                )

            full_path = destination
            if "path" in data and data["path"]:
                full_path = fileManager.join_path(target, data["path"], destination)
            else:
                path, _ = fileManager.split_path(target, filename)
                if path:
                    full_path = fileManager.join_path(target, path, destination)

            canon_path, canon_name = fileManager.canonicalize(target, full_path)
            sanitized_name = fileManager.sanitize_name(target, canon_name)

            if canon_path:
                full_path = fileManager.join_path(target, canon_path, sanitized_name)
            else:
                full_path = sanitized_name

            # prohibit overwriting the file that is currently being printed
            currentOrigin, currentFilename = _getCurrentFile()
            if (
                currentFilename == full_path
                and currentOrigin == target
                and (printer.is_printing() or printer.is_paused())
            ):
                make_response(
                    "Trying to slice into file that is currently being printed: %s"
                    % full_path,
                    409,
                )

            if "profile" in data and data["profile"]:
                profile = data["profile"]
                del data["profile"]
            else:
                profile = None

            if "printerProfile" in data and data["printerProfile"]:
                printerProfile = data["printerProfile"]
                del data["printerProfile"]
            else:
                printerProfile = None

            if (
                "position" in data
                and data["position"]
                and isinstance(data["position"], dict)
                and "x" in data["position"]
                and "y" in data["position"]
            ):
                position = data["position"]
                del data["position"]
            else:
                position = None

            select_after_slicing = False
            if "select" in data and data["select"] in valid_boolean_trues:
                if not printer.is_operational():
                    return make_response(
                        "Printer is not operational, cannot directly select for printing",
                        409,
                    )
                select_after_slicing = True

            print_after_slicing = False
            if "print" in data and data["print"] in valid_boolean_trues:
                if not printer.is_operational():
                    return make_response(
                        "Printer is not operational, cannot directly start printing", 409
                    )
                select_after_slicing = print_after_slicing = True

            override_keys = [
                k for k in data if k.startswith("profile.") and data[k] is not None
            ]
            overrides = {}
            for key in override_keys:
                overrides[key[len("profile.") :]] = data[key]

            def slicing_done(target, path, select_after_slicing, print_after_slicing):
                if select_after_slicing or print_after_slicing:
                    sd = False
                    if target == FileDestinations.SDCARD:
                        filenameToSelect = path
                        sd = True
                    else:
                        filenameToSelect = fileManager.path_on_disk(target, path)
                    printer.select_file(filenameToSelect, sd, print_after_slicing, user)

            try:
                fileManager.slice(
                    slicer,
                    target,
                    filename,
                    target,
                    full_path,
                    profile=profile,
                    printer_profile_id=printerProfile,
                    position=position,
                    overrides=overrides,
                    display=canon_name,
                    callback=slicing_done,
                    callback_args=(
                        target,
                        full_path,
                        select_after_slicing,
                        print_after_slicing,
                    ),
                )
            except octoprint.slicing.UnknownProfile:
                return make_response(
                    "Profile {profile} doesn't exist".format(**locals()), 400
                )

            files = {}
            location = url_for(
                ".readGcodeFile", target=target, filename=full_path, _external=True
            )
            result = {
                "name": destination,
                "path": full_path,
                "display": canon_name,
                "origin": FileDestinations.LOCAL,
                "refs": {
                    "resource": location,
                    "download": url_for("index", _external=True)
                    + "downloads/files/"
                    + target
                    + "/"
                    + full_path,
                },
            }

            r = make_response(jsonify(result), 202)
            r.headers["Location"] = location
            return r

    elif command == "analyse":
        with Permissions.FILES_UPLOAD.require(403):
            if not _verifyFileExists(target, filename):
                return make_response(
                    "File not found on '%s': %s" % (target, filename), 404
                )

            printer_profile = None
            if "printerProfile" in data and data["printerProfile"]:
                printer_profile = data["printerProfile"]

            if not fileManager.analyse(
                target, filename, printer_profile_id=printer_profile
            ):
                return make_response(
                    "No analysis possible for {} on {}".format(filename, target), 400
                )

    elif command == "copy" or command == "move":
        with Permissions.FILES_UPLOAD.require(403):
            # Copy and move are only possible on local storage
            if target not in [FileDestinations.LOCAL]:
                return make_response(
                    "Unsupported target for {}: {}".format(command, target), 400
                )

            if not _verifyFileExists(target, filename) and not _verifyFolderExists(
                target, filename
            ):
                return make_response(
                    "File or folder not found on {}: {}".format(target, filename), 404
                )

            path, name = fileManager.split_path(target, filename)

            destination = data["destination"]
            dst_path, dst_name = fileManager.split_path(target, destination)
            sanitized_destination = fileManager.join_path(
                target, dst_path, fileManager.sanitize_name(target, dst_name)
            )

            if (
                _verifyFolderExists(target, destination)
                and sanitized_destination != filename
            ):
                # destination is an existing folder and not ourselves (= display rename), we'll assume we are supposed
                # to move filename to this folder under the same name
                destination = fileManager.join_path(target, destination, name)

            if _verifyFileExists(target, destination) or _verifyFolderExists(
                target, destination
            ):
                return make_response(
                    "File or folder does already exist on {}: {}".format(
                        target, destination
                    ),
                    409,
                )

            is_file = fileManager.file_exists(target, filename)
            is_folder = fileManager.folder_exists(target, filename)

            if not (is_file or is_folder):
                return make_response(
                    "{} on {} is neither file or folder, can't {}".format(
                        filename, target, command
                    ),
                    400,
                )

            if command == "copy":
                # destination already there? error...
                if _verifyFileExists(target, destination) or _verifyFolderExists(
                    target, destination
                ):
                    return make_response(
                        "File or folder does already exist on {}: {}".format(
                            target, destination
                        ),
                        409,
                    )

                if is_file:
                    fileManager.copy_file(target, filename, destination)
                else:
                    fileManager.copy_folder(target, filename, destination)

            elif command == "move":
                with Permissions.FILES_DELETE.require(403):
                    if _isBusy(target, filename):
                        return make_response(
                            "Trying to move a file or folder that is currently in use: {}".format(
                                filename
                            ),
                            409,
                        )

                    # destination already there AND not ourselves (= display rename)? error...
                    if (
                        _verifyFileExists(target, destination)
                        or _verifyFolderExists(target, destination)
                    ) and sanitized_destination != filename:
                        return make_response(
                            "File or folder does already exist on {}: {}".format(
                                target, destination
                            ),
                            409,
                        )

                    # deselect the file if it's currently selected
                    currentOrigin, currentFilename = _getCurrentFile()
                    if currentFilename is not None and filename == currentFilename:
                        printer.unselect_file()

                    if is_file:
                        fileManager.move_file(target, filename, destination)
                    else:
                        fileManager.move_folder(target, filename, destination)

            location = url_for(
                ".readGcodeFile", target=target, filename=destination, _external=True
            )
            result = {
                "name": name,
                "path": destination,
                "origin": FileDestinations.LOCAL,
                "refs": {"resource": location},
            }
            if is_file:
                result["refs"]["download"] = (
                    url_for("index", _external=True)
                    + "downloads/files/"
                    + target
                    + "/"
                    + destination
                )

            r = make_response(jsonify(result), 201)
            r.headers["Location"] = location
            return r

    return NO_CONTENT


@api.route("/files/<string:target>/<path:filename>", methods=["DELETE"])
@no_firstrun_access
@Permissions.FILES_DELETE.require(403)
def deleteGcodeFile(filename, target):
    if not _verifyFileExists(target, filename) and not _verifyFolderExists(
        target, filename
    ):
        return make_response(
            "File/Folder not found on '%s': %s" % (target, filename), 404
        )

    if _verifyFileExists(target, filename):
        if target not in [FileDestinations.LOCAL, FileDestinations.SDCARD]:
            return make_response("Unknown target: %s" % target, 400)

        if _isBusy(target, filename):
            return make_response(
                "Trying to delete a file that is currently in use: %s" % filename, 409
            )

        # deselect the file if it's currently selected
        currentOrigin, currentPath = _getCurrentFile()
        if (
            currentPath is not None
            and currentOrigin == target
            and filename == currentPath
        ):
            printer.unselect_file()

        # delete it
        if target == FileDestinations.SDCARD:
            printer.delete_sd_file(filename, tags={"source:api", "api:files.sd"})
        else:
            fileManager.remove_file(target, filename)

    elif _verifyFolderExists(target, filename):
        if target not in [FileDestinations.LOCAL]:
            return make_response("Unknown target: %s" % target, 400)

        if _isBusy(target, filename):
            return make_response(
                "Trying to delete a folder that contains a file that is currently in use: %s"
                % filename,
                409,
            )

        # deselect the file if it's currently selected
        currentOrigin, currentPath = _getCurrentFile()
        if (
            currentPath is not None
            and currentOrigin == target
            and fileManager.file_in_path(target, filename, currentPath)
        ):
            printer.unselect_file()

        # delete it
        fileManager.remove_folder(target, filename, recursive=True)

    return NO_CONTENT


def _getCurrentFile():
    currentJob = printer.get_current_job()
    if (
        currentJob is not None
        and "file" in currentJob
        and "path" in currentJob["file"]
        and "origin" in currentJob["file"]
    ):
        return currentJob["file"]["origin"], currentJob["file"]["path"]
    else:
        return None, None


class WerkzeugFileWrapper(octoprint.filemanager.util.AbstractFileWrapper):
    """
    A wrapper around a Werkzeug ``FileStorage`` object.

    Arguments:
        file_obj (werkzeug.datastructures.FileStorage): The Werkzeug ``FileStorage`` instance to wrap.

    .. seealso::

       `werkzeug.datastructures.FileStorage <http://werkzeug.pocoo.org/docs/0.10/datastructures/#werkzeug.datastructures.FileStorage>`_
            The documentation of Werkzeug's ``FileStorage`` class.
    """

    def __init__(self, file_obj):
        octoprint.filemanager.util.AbstractFileWrapper.__init__(self, file_obj.filename)
        self.file_obj = file_obj

    def save(self, path):
        """
        Delegates to ``werkzeug.datastructures.FileStorage.save``
        """
        self.file_obj.save(path)

    def stream(self):
        """
        Returns ``werkzeug.datastructures.FileStorage.stream``
        """
        return self.file_obj.stream



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"

from flask_babel import gettext

import octoprint.plugin

from .subwizards import Subwizards


class CoreWizardPlugin(
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.WizardPlugin,
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.BlueprintPlugin,
    Subwizards,
):

    # ~~ TemplatePlugin API

    def get_template_configs(self):
        required = self._get_subwizard_attrs("_is_", "_wizard_required")
        names = self._get_subwizard_attrs("_get_", "_wizard_name")
        additional = self._get_subwizard_attrs(
            "_get_", "_additional_wizard_template_data"
        )

        firstrunonly = self._get_subwizard_attrs("_is_", "_wizard_firstrunonly")
        firstrun = self._settings.global_get(["server", "firstRun"])

        if not firstrun:
            required = dict(
                (key, value)
                for key, value in required.items()
                if not firstrunonly.get(key, lambda: False)()
            )

        result = list()
        for key, method in required.items():
            if not callable(method):
                continue

            if not method():
                continue

            if key not in names:
                continue

            name = names[key]()
            if not name:
                continue

            config = {
                "type": "wizard",
                "name": name,
                "template": "corewizard_{}_wizard.jinja2".format(key),
                "div": "wizard_plugin_corewizard_{}".format(key),
                "suffix": "_{}".format(key),
            }
            if key in additional:
                additional_result = additional[key]()
                if additional_result:
                    config.update(additional_result)
            result.append(config)

        return result

    # ~~ AssetPlugin API

    def get_assets(self):
        if self.is_wizard_required():
            return {"js": ["js/corewizard.js"], "css": ["css/corewizard.css"]}
        else:
            return {}

    # ~~ WizardPlugin API

    def is_wizard_required(self):
        required = self._get_subwizard_attrs("_is_", "_wizard_required")
        firstrunonly = self._get_subwizard_attrs("_is_", "_wizard_firstrunonly")
        firstrun = self._settings.global_get(["server", "firstRun"])

        if not firstrun:
            required = dict(
                (key, value)
                for key, value in required.items()
                if not firstrunonly.get(key, lambda: False)()
            )
        any_required = any(map(lambda m: m(), required.values()))

        return any_required

    def get_wizard_details(self):
        result = {}

        def add_result(key, method):
            result[key] = method()

        self._get_subwizard_attrs("_get_", "_wizard_details", add_result)

        return result

    def get_wizard_version(self):
        return 4

    # ~~ helpers

    def _get_subwizard_attrs(self, start, end, callback=None):
        result = {}

        for item in dir(self):
            if not item.startswith(start) or not item.endswith(end):
                continue

            key = item[len(start) : -len(end)]
            if not key:
                continue

            attr = getattr(self, item)
            if callable(callback):
                callback(key, attr)
            result[key] = attr

        return result


__plugin_name__ = "Core Wizard"
__plugin_author__ = "Gina Häußge"
__plugin_description__ = "Provides wizard dialogs for core components and functionality"
__plugin_disabling_discouraged__ = gettext(
    "Without this plugin OctoPrint will no longer be able to perform "
    "setup steps that might be required after an update."
)
__plugin_license__ = "AGPLv3"
__plugin_pythoncompat__ = ">=2.7,<4"
__plugin_implementation__ = CoreWizardPlugin()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2018 The OctoPrint Project - Released under terms of the AGPLv3 License"


def get_latest(target, check, online=True, *args, **kwargs):
    current_version = check.get("current_version", "1.0.0")

    information = {
        "local": {"name": current_version, "value": current_version},
        "remote": {"name": current_version, "value": current_version},
    }

    return information, True



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from ..exceptions import CannotCheckOffline, ConfigurationInvalid


def get_latest(target, check, full_data=False, online=True, *args, **kwargs):
    python_checker = check.get("python_checker")
    if python_checker is None or not hasattr(python_checker, "get_latest"):
        raise ConfigurationInvalid(
            'Update configuration for {} of type python_checker needs python_checker defined and have an attribute "get_latest"'.format(
                target
            )
        )

    if not online and not check.get("offline", False):
        raise CannotCheckOffline(
            "{} isn't marked as 'offline' capable, but we are apparently offline right now".format(
                target
            )
        )

    try:
        return check["python_checker"].get_latest(
            target, check, full_data=full_data, online=online
        )
    except Exception:
        import inspect

        args, _, _, _ = inspect.getargspec(check["python_checker"].get_latest)
        if "online" not in args:
            # old python_checker footprint, simply leave out the online parameter
            return check["python_checker"].get_latest(target, check, full_data=full_data)

        # some other error, raise again
        raise
